@@ -1,9 +1,9 @@
 #
-# Shorewall-perl 4.0 -- /usr/share/shorewall-perl/Shorewall/Accounting.pm
+# Shorewall-perl 4.2 -- /usr/share/shorewall-perl/Shorewall/Accounting.pm
 #
 #     This program is under GPL [http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt]
 #
-#     (c) 2007 - Tom Eastep (teastep@shorewall.net)
+#     (c) 2007,2008 - Tom Eastep (teastep@shorewall.net)
 #
 #       Complete documentation is available at http://shorewall.net
 #
@@ -35,7 +35,7 @@ use strict;
 our @ISA = qw(Exporter);
 our @EXPORT = qw( setup_accounting );
 our @EXPORT_OK = qw( );
-our $VERSION = 4.0.6;
+our $VERSION = 4.2.4;
 
 #
 # Initialize globals -- we take this novel approach to globals initialization to allow
@@ -64,9 +64,11 @@ sub process_accounting_rule( $$$$$$$$$ ) {
 
     my ($action, $chain, $source, $dest, $proto, $ports, $sports, $user, $mark ) = @_;
 
-    sub check_for_builtin( $ ) {
+    our $disposition = '';
+
+    sub check_chain( $ ) {
 	my $chainref = shift;
-	fatal_error "A builtin Chain ($jumpchainref->{name}) may not appear in the accounting file" if $chainref->{builtin};
+	fatal_error "A non-accounting chain ($chainref->{name}) may not appear in the accounting file" if $chainref->{policy};
     }
 
     sub accounting_error() {
@@ -75,9 +77,9 @@ sub process_accounting_rule( $$$$$$$$$ ) {
 
     sub jump_to_chain( $ ) {
 	my $jumpchain = $_[0];
-	$jumpchainref = ensure_chain( 'filter', $jumpchain );
-	check_for_builtin( $jumpchainref );
-	mark_referenced $jumpchainref;
+	$jumpchainref = ensure_accounting_chain( $jumpchain );
+	check_chain( $jumpchainref );
+	$disposition = $jumpchain;
 	"-j $jumpchain";
     }
 
@@ -98,55 +100,51 @@ sub process_accounting_rule( $$$$$$$$$ ) {
 	    if ( $cmd ) {
 		if ( $cmd eq 'COUNT' ) {
 		    $rule2=1;
-		    $target = jump_to_chain $action;
-		} elsif ( $cmd eq 'JUMP' ) {
-		    $target = jump_to_chain $action;
-		} else {
+		} elsif ( $cmd ne 'JUMP' ) {
 		    accounting_error;
 		}
-	    } else {
-		$target = jump_to_chain $action;
-	    }
+	    } 
+
+	    $target = jump_to_chain $action;
 	}
     }
 
     my $restriction = NO_RESTRICT;
 
-    $source = ALLIPv4 if $source eq 'any' || $source eq 'all';
+    $source = ALLIP if $source eq 'any' || $source eq 'all';
 
     if ( have_bridges ) {
 	my $fw = firewall_zone;
 
 	if ( $source =~ /^$fw:?(.*)$/ ) {
-	    $source = $1 ? $1 : ALLIPv4;
+	    $source = $1 ? $1 : ALLIP;
 	    $restriction = OUTPUT_RESTRICT;
 	    $chain = 'accountout' unless $chain and $chain ne '-';
-	    $dest = ALLIPv4 if $dest   eq 'any' || $dest   eq 'all';
+	    $dest = ALLIP if $dest   eq 'any' || $dest   eq 'all';
 	} else {
 	    $chain = 'accounting' unless $chain and $chain ne '-';
-	    if ( $dest eq 'any' || $dest eq 'all' || $dest eq ALLIPv4 ) {
+	    if ( $dest eq 'any' || $dest eq 'all' || $dest eq ALLIP ) {
 		expand_rule(
 			    ensure_filter_chain( 'accountout' , 0 ) ,
 			    OUTPUT_RESTRICT ,
 			    $rule ,
 			    $source ,
-			    $dest = ALLIPv4 ,
+			    $dest = ALLIP ,
+			    '' ,
 			    '' ,
 			    $target ,
 			    '' ,
-			    '' ,
+			    $disposition ,
 			    ''  );
 	    }
 	}
     } else {
 	$chain = 'accounting' unless $chain and $chain ne '-';
-	$dest = ALLIPv4 if $dest   eq 'any' || $dest   eq 'all';
+	$dest = ALLIP if $dest   eq 'any' || $dest   eq 'all';
     }
 
-    my $chainref = ensure_filter_chain $chain , 0;
+    my $chainref = ensure_accounting_chain $chain;
 
-    check_for_builtin( $chainref );
-    
     expand_rule
 	$chainref ,
 	$restriction ,
@@ -154,9 +152,10 @@ sub process_accounting_rule( $$$$$$$$$ ) {
 	$source ,
 	$dest ,
 	'' ,
+	'' ,
 	$target ,
 	'' ,
-	'' ,
+	$disposition ,
 	'' ;
 
     if ( $rule2 ) {
@@ -170,6 +169,7 @@ sub process_accounting_rule( $$$$$$$$$ ) {
 	    '' ,
 	    '' ,
 	    '' ,
+	    '' ,
 	    '' ;
     }
 }
@@ -180,6 +180,8 @@ sub setup_accounting() {
 
     first_entry "$doing $fn...";
 
+    my $nonEmpty = 0;
+
     while ( read_a_line ) {
 
 	my ( $action, $chain, $source, $dest, $proto, $ports, $sports, $user, $mark ) = split_line1 1, 9, 'Accounting File';
@@ -187,26 +189,29 @@ sub setup_accounting() {
 	if ( $action eq 'COMMENT' ) {
 	    process_comment;
 	} else {
+	    $nonEmpty = 1;
 	    process_accounting_rule $action, $chain, $source, $dest, $proto, $ports, $sports, $user, $mark;
 	}
     }
 
+    fatal_error "Accounring rules are isolated" if $nonEmpty && ! $filter_table->{accounting};
+
     clear_comment;
 
     if ( have_bridges ) {
 	if ( $filter_table->{accounting} ) {
 	    for my $chain ( qw/INPUT FORWARD/ ) {
-		insert_rule $filter_table->{$chain}, 1, '-j accounting';
+		insert_rule1 $filter_table->{$chain}, 0, '-j accounting';
 	    }
 	}
 
 	if ( $filter_table->{accountout} ) {
-	    insert_rule $filter_table->{OUTPUT}, 1, '-j accountout';
+	    insert_rule1 $filter_table->{OUTPUT}, 0, '-j accountout';
 	}
     } else {
 	if ( $filter_table->{accounting} ) {
 	    for my $chain ( qw/INPUT FORWARD OUTPUT/ ) {
-		insert_rule $filter_table->{$chain}, 1, '-j accounting';
+		insert_rule1 $filter_table->{$chain}, 0, '-j accounting';
 	    }
 	}
     }
