@@ -1,9 +1,9 @@
 #
-# Shorewall-perl 4.0 -- /usr/share/shorewall-perl/Shorewall/Actions.pm
+# Shorewall-perl 4.2 -- /usr/share/shorewall-perl/Shorewall/Actions.pm
 #
 #     This program is under GPL [http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt]
 #
-#     (c) 2007 - Tom Eastep (teastep@shorewall.net)
+#     (c) 2007,2008 - Tom Eastep (teastep@shorewall.net)
 #
 #       Complete documentation is available at http://shorewall.net
 #
@@ -34,6 +34,7 @@ use strict;
 our @ISA = qw(Exporter);
 our @EXPORT = qw( merge_levels
 		  isolate_basic_target
+		  get_target_param
 		  add_requiredby
 		  createactionchain
 		  find_logactionchain
@@ -52,9 +53,10 @@ our @EXPORT = qw( merge_levels
 		  %actions
 
 		  %macros
+		  $macro_commands
 		  );
 our @EXPORT_OK = qw( initialize );
-our $VERSION = 4.0.6;
+our $VERSION = 4.2.4;
 
 #
 #  Used Actions. Each action that is actually used has an entry with value 1.
@@ -81,6 +83,13 @@ our %logactionchains;
 
 our %macros;
 
+our $family;
+
+#
+# Commands that can be embedded in a macro file and how many total tokens on the line (0 => unlimited).
+#
+our $macro_commands = { COMMENT => 0, FORMAT => 2 };
+
 #
 # Initialize globals -- we take this novel approach to globals initialization to allow
 #                       the compiler to run multiple times in the same process. The
@@ -90,7 +99,9 @@ our %macros;
 #                       the second and subsequent calls to that function.
 #
 
-sub initialize() {
+sub initialize( $ ) {
+
+    $family          = shift;
     %usedactions     = ();
     %default_actions = ( DROP     => 'none' ,
 			 REJECT   => 'none' ,
@@ -98,10 +109,11 @@ sub initialize() {
 			 QUEUE    => 'none' );
     %actions         = ();
     %logactionchains = ();
+    %macros          = ();
 }
 
 INIT {
-    initialize;
+    initialize( F_IPV4 );
 }
 
 #
@@ -219,7 +231,22 @@ sub merge_macro_column( $$ ) {
 # Get Macro Name -- strips away trailing /* and :* from the first column in a rule, macro or action.
 #
 sub isolate_basic_target( $ ) {
-    ( split '[/:]', $_[0])[0];
+    my $target = ( split '[/:]', $_[0])[0]; 
+
+    $target =~ /^(\w+)[(].*[)]$/ ? $1 : $target;
+}
+
+#
+# Split the passed target into the basic target and parameter
+#
+sub get_target_param( $ ) {
+    my ( $target, $param ) = split '/', $_[0];
+
+    unless ( defined $param ) {
+	( $target, $param ) = ( $1, $2 ) if $target =~ /^(.*?)[(](.*)[)]$/;
+    }
+
+    ( $target, $param );
 }
 
 #
@@ -259,7 +286,7 @@ sub createlogactionchain( $$ ) {
     my $chain = $action;
     my $actionref = $actions{$action};
     my $chainref;
-    
+
     my ($lev, $tag) = split ':', $level;
 
     validate_level $lev;
@@ -277,8 +304,6 @@ sub createlogactionchain( $$ ) {
     $logactionchains{"$action:$level"} = $chainref = new_standard_chain '%' . $chain . $actionref->{actchain}++;
 
     fatal_error "Too many invocations of Action $action" if $actionref->{actchain} > 99;
-    
-    mark_referenced $chainref; # Just in case the action body is empty.
 
     unless ( $targets{$action} & STANDARD ) {
 
@@ -306,8 +331,6 @@ sub createsimpleactionchain( $ ) {
 
     $logactionchains{"$action:none"} = $chainref;
 
-    mark_referenced $chainref; # Just in case the action body is empty.
-
     unless ( $targets{$action} & STANDARD ) {
 
 	my $file = find_file $action;
@@ -389,7 +412,9 @@ sub process_macro1 ( $$ ) {
     push_open( $macrofile );
 
     while ( read_a_line ) {
-	my ( $mtarget, $msource,  $mdest,  $mproto,  $mports,  $msports, $mrate, $muser ) = split_line 1, 8, 'macro file';
+	my ( $mtarget, $msource,  $mdest,  $mproto,  $mports,  $msports, $morigdest, $mrate, $muser ) = split_line1 1, 9, 'macro file', $macro_commands;
+
+	next if $mtarget eq 'COMMENT' || $mtarget eq 'FORMAT';
 
 	$mtarget =~ s/:.*$//;
 
@@ -418,7 +443,7 @@ sub process_action1 ( $$ ) {
     my $targettype = $targets{$target};
 
     if ( defined $targettype ) {
-	return if ( $targettype == STANDARD ) || ( $targettype == MACRO ) || ( $targettype & ( LOGRULE |  NFQ | CHAIN ) );
+	return if ( $targettype == STANDARD ) || ( $targettype & ( MACRO | LOGRULE |  NFQ | CHAIN ) );
 
 	fatal_error "Invalid TARGET ($target)" if $targettype & STANDARD;
 
@@ -428,7 +453,7 @@ sub process_action1 ( $$ ) {
     } elsif ( $target eq 'COMMENT' ) {
 	fatal_error "Invalid TARGET ($wholetarget)" unless $wholetarget eq $target;
     } else {
-	( $target, my $param ) = split '/', $target;
+	( $target, my $param ) = get_target_param $target;
 
 	return if $target eq 'NFQUEUE';
 
@@ -532,39 +557,79 @@ sub process_action( $$$$$$$$$$ ) {
 
     my ( $action , $level ) = split_action $target;
 
+    if ( $action eq 'REJECT' ) {
+	$action = 'reject';
+    } elsif ( $action eq 'CONTINUE' ) {
+	$action = 'RETURN';
+    } elsif ( $action =~ /^NFQUEUE/ ) {
+	( $action, my $param ) = get_target_param $action;
+	$param = 1 unless defined $param;
+	$action = "NFQUEUE --queue-num $param";
+    } elsif ( $action eq 'COUNT' ) {
+	$action = '';
+    }
+
     expand_rule ( $chainref ,
 		  NO_RESTRICT ,
 		  do_proto( $proto, $ports, $sports ) . do_ratelimit( $rate, $action ) . do_user $user ,
 		  $source ,
 		  $dest ,
 		  '', #Original Dest
-		  '-j ' . ($action eq 'REJECT' ? 'reject' : $action eq 'CONTINUE' ? 'RETURN' : $action),
+		  '', #Original Dest port
+		  $action ? "-j $action" : '',
 		  $level ,
 		  $action ,
 		  '' );
 }
 
 #
-# Expand Macro in action file4s.
+# Expand Macro in action files.
 #
 sub process_macro3( $$$$$$$$$$$ ) {
-    my ( $fn, $param, $chainref, $action, $source, $dest, $proto, $ports, $sports, $rate, $user ) = @_;
+    my ( $macro, $param, $chainref, $action, $source, $dest, $proto, $ports, $sports, $rate, $user ) = @_;
+
+    my $nocomment = no_comment;
+
+    my $format = 1;
+
+    macro_comment $macro;
+
+    my $fn = $macros{$macro};
 
     progress_message "..Expanding Macro $fn...";
 
     push_open $fn;
 
-    my $standard = ( $fn =~ /^($globals{SHAREDIR})/ );
-
     while ( read_a_line ) {
 
-	my ( $mtarget, $msource, $mdest, $mproto, $mports, $msports, $mrate, $muser ) = split_line 1, 8, 'macro file';
+	my ( $mtarget, $msource, $mdest, $mproto, $mports, $msports, $morigdest, $mrate, $muser );
+
+	if ( $format == 1 ) {
+	    ( $mtarget, $msource, $mdest, $mproto, $mports, $msports, $mrate, $muser, $morigdest ) = split_line1 1, 9, 'macro file', $macro_commands;
+	} else {
+	    ( $mtarget, $msource, $mdest, $mproto, $mports, $msports, $morigdest, $mrate, $muser ) = split_line1 1, 9, 'macro file', $macro_commands;
+	}
+
+	if ( $mtarget eq 'COMMENT' ) {
+	    process_comment unless $nocomment;
+	    next;
+	}
+
+	if ( $mtarget eq 'FORMAT' ) {
+	    fatal_error "Invalid FORMAT ($msource)" unless $msource =~ /^[12]$/;
+	    $format = $msource;
+	    next;
+	}
+
+	fatal_error "Invalid macro file entry (too many columns)" if $morigdest ne '-' && $format == 1;
 
 	if ( $mtarget =~ /^PARAM:?/ ) {
 	    fatal_error 'PARAM requires that a parameter be supplied in macro invocation' unless $param;
 	    $mtarget = substitute_param $param,  $mtarget;
 	}
 
+	fatal_error "Macros used within Actions may not specify an ORIGINAL DEST " if $morigdest ne '-';
+
 	if ( $msource ) {
 	    if ( ( $msource eq '-' ) || ( $msource eq 'SOURCE' ) ) {
 		$msource = $source || '';
@@ -604,7 +669,9 @@ sub process_macro3( $$$$$$$$$$$ ) {
 
     pop_open;
 
-    progress_message '..End Macro'
+    progress_message '..End Macro';
+
+    clear_comment unless $nocomment;
 }
 
 #
@@ -613,7 +680,6 @@ sub process_macro3( $$$$$$$$$$$ ) {
 sub process_action3( $$$$$ ) {
     my ( $chainref, $wholeaction, $action, $level, $tag ) = @_;
     my $actionfile = find_file "action.$action";
-    my $standard = ( $actionfile =~ /^$globals{SHAREDIR}/ );
 
     fatal_error "Missing Action File ($actionfile)" unless -f $actionfile;
 
@@ -634,22 +700,20 @@ sub process_action3( $$$$$ ) {
 
 	my ( $action2 , $level2 ) = split_action $target2;
 
-	my $action2type = $targets{isolate_basic_target $action2};
+	( $action2 , my $param ) = get_target_param $action2;
+
+	my $action2type = $targets{$action2} || 0; 
 
 	unless ( $action2type == STANDARD ) {
 	    if ( $action2type & ACTION ) {
 		$target2 = (find_logactionchain ( $target = $target2 ))->{name};
 	    } else {
-		fatal_error "Internal Error" unless $action2type == MACRO || $action2type & ( LOGRULE | NFQ | CHAIN );
+		fatal_error "Internal Error" unless $action2type & ( MACRO | LOGRULE | NFQ | CHAIN );
 	    }
 	}
 
 	if ( $action2type == MACRO ) {
-	    ( $action2, my $param ) = split '/', $action2;
-
-	    fatal_error "Null Macro" unless my $fn = $macros{$action2};
-
-	    process_macro3( $fn, $param, $chainref, $action, $source, $dest, $proto, $ports, $sports, $rate, $user );
+	    process_macro3( $action2, $param, $chainref, $action, $source, $dest, $proto, $ports, $sports, $rate, $user );
 	} else {
 	    process_action $chainref, $action, $target2, $source, $dest, $proto, $ports, $sports, $rate, $user;
 	}
@@ -673,7 +737,12 @@ sub process_actions3 () {
 
 	    add_rule $chainref, '-m addrtype --dst-type BROADCAST -j DROP';
 	} else {
-	    add_command $chainref, 'for address in $ALL_BCASTS; do';
+	    if ( $family == F_IPV4 ) {
+		add_command $chainref, 'for address in $ALL_BCASTS; do';
+	    } else {
+		add_command $chainref, 'for address in $ALL_ACASTS; do';
+	    }
+
 	    incr_cmd_level $chainref;
 	    log_rule_limit $level, $chainref, 'dropBcast' , 'DROP', '', $tag, 'add', ' -d $address ' if $level ne '';
 	    add_rule $chainref, '-d $address -j DROP';
@@ -683,30 +752,46 @@ sub process_actions3 () {
 	    log_rule_limit $level, $chainref, 'dropBcast' , 'DROP', '', $tag, 'add', ' -d 224.0.0.0/4 ' if $level ne '';
 	}
 
-	add_rule $chainref, '-d 224.0.0.0/4 -j DROP';
+
+	if ( $family == F_IPV4 ) {
+	    add_rule $chainref, '-d 224.0.0.0/4 -j DROP';
+	} else {
+	    add_rule $chainref, '-d ff00::/10 -j DROP';
+	}
     }
 
     sub allowBcast( $$$ ) {
 	my ($chainref, $level, $tag) = @_;
 
-	if ( $capabilities{ADDRTYPE} ) {
+	if ( $family == F_IPV4 && $capabilities{ADDRTYPE} ) {
 	    if ( $level ne '' ) {
 		log_rule_limit $level, $chainref, 'allowBcast' , 'ACCEPT', '', $tag, 'add', ' -m addrtype --dst-type BROADCAST ';
 		log_rule_limit $level, $chainref, 'allowBcast' , 'ACCEPT', '', $tag, 'add', ' -d 224.0.0.0/4 ';
 	    }
 
 	    add_rule $chainref, '-m addrtype --dst-type BROADCAST -j ACCEPT';
+	    add_rule $chainref, '-d 224.0.0.0/4 -j ACCEPT';
 	} else {
-	    add_command $chainref, 'for address in $ALL_BCASTS; do';
+	    if ( $family == F_IPV4 ) {
+		add_command $chainref, 'for address in $ALL_BCASTS; do';
+	    } else {
+		add_command $chainref, 'for address in $ALL_MACASTS; do';
+	    }
+
 	    incr_cmd_level $chainref;
 	    log_rule_limit $level, $chainref, 'allowBcast' , 'ACCEPT', '', $tag, 'add', ' -d $address ' if $level ne '';
 	    add_rule $chainref, '-d $address -j ACCEPT';
 	    decr_cmd_level $chainref;
 	    add_command $chainref, 'done';
 
-	    log_rule_limit $level, $chainref, 'allowBcast' , 'ACCEPT', '', $tag, 'add', ' -d 224.0.0.0/4 ' if $level ne '';
+	    if ( $family == F_IPV4 ) {
+		log_rule_limit $level, $chainref, 'allowBcast' , 'ACCEPT', '', $tag, 'add', ' -d 224.0.0.0/4 ' if $level ne '';
+		add_rule $chainref, '-d 224.0.0.0/4 -j ACCEPT';
+	    } else {
+		log_rule_limit $level, $chainref, 'allowBcast' , 'ACCEPT', '', $tag, 'add', ' -d ff00::/10 ' if $level ne '';
+		add_rule $chainref, '-d ff00:/10 -j ACCEPT';
+	    }
 	}
-	    add_rule $chainref, '-d 224.0.0.0/4 -j ACCEPT';
     }
 
     sub dropNotSyn ( $$$ ) {
@@ -783,16 +868,15 @@ sub process_actions3 () {
 	add_rule $chainref, '-j ACCEPT';
     }
 
-    my %builtinops = ( 'dropBcast'    => \&dropBcast,
-		       'allowBcast'   => \&allowBcast,
-		       'dropNotSyn'   => \&dropNotSyn,
-		       'rejNotSyn'    => \&rejNotSyn,
-		       'dropInvalid'  => \&dropInvalid,
-		       'allowInvalid' => \&allowInvalid,
-		       'allowinUPnP'  => \&allowinUPnP,
-		       'forwardUPnP'  => \&forwardUPnP,
-		       'Limit'        => \&Limit,
-		       );
+    my %builtinops = ( 'dropBcast'      => \&dropBcast,
+		       'allowBcast'     => \&allowBcast,
+		       'dropNotSyn'     => \&dropNotSyn,
+		       'rejNotSyn'      => \&rejNotSyn,
+		       'dropInvalid'    => \&dropInvalid, 
+		       'allowInvalid'   => \&allowInvalid,
+		       'allowinUPnP'    => \&allowinUPnP, 
+		       'forwardUPnP'    => \&forwardUPnP, 
+		       'Limit'          => \&Limit, );
 
     for my $wholeaction ( keys %usedactions ) {
 	my $chainref = find_logactionchain $wholeaction;
