@@ -1,9 +1,9 @@
 #
-# Shorewall-perl 4.0 -- /usr/share/shorewall-perl/Shorewall/Nat.pm
+# Shorewall-perl 4.2 -- /usr/share/shorewall-perl/Shorewall/Nat.pm
 #
 #     This program is under GPL [http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt]
 #
-#     (c) 2007 - Tom Eastep (teastep@shorewall.net)
+#     (c) 2007,2008 - Tom Eastep (teastep@shorewall.net)
 #
 #       Complete documentation is available at http://shorewall.net
 #
@@ -30,13 +30,14 @@ use Shorewall::IPAddrs;
 use Shorewall::Zones;
 use Shorewall::Chains qw(:DEFAULT :internal);
 use Shorewall::IPAddrs;
+use Shorewall::Providers qw( lookup_provider );
 
 use strict;
 
 our @ISA = qw(Exporter);
 our @EXPORT = qw( setup_masq setup_nat setup_netmap add_addresses );
 our @EXPORT_OK = ();
-our $VERSION = 4.0.6;
+our $VERSION = 4.2.4;
 
 our @addresses_to_add;
 our %addresses_to_add;
@@ -77,7 +78,7 @@ sub do_ipsec_options($)
     my $options = '-m policy --pol ipsec --dir out ';
     my $fmt;
 
-    for my $e ( split ',' , $list ) {
+    for my $e ( split_list $list, 'option' ) {
 	my $val    = undef;
 	my $invert = '';
 
@@ -114,13 +115,42 @@ sub do_ipsec_options($)
 #
 sub setup_one_masq($$$$$$$)
 {
-    my ($fullinterface, $networks, $addresses, $proto, $ports, $ipsec, $mark) = @_;
+    my ($interfacelist, $networks, $addresses, $proto, $ports, $ipsec, $mark) = @_;
 
-    my $rule = '';
     my $pre_nat;
     my $add_snat_aliases = $config{ADD_SNAT_ALIASES};
     my $destnets = '';
-    my $target = '-j MASQUERADE ';
+    my $baserule = '';
+
+    #
+    # Leading '+'
+    #
+    $pre_nat = 1 if $interfacelist =~ s/^\+//;
+    #
+    # Parse the remaining part of the INTERFACE column
+    #
+    if ( $interfacelist =~ /^([^:]+)::([^:]*)$/ ) {
+	$add_snat_aliases = 0;
+	$destnets = $2;
+	$interfacelist = $1;
+    } elsif ( $interfacelist =~ /^([^:]+:[^:]+):([^:]+)$/ ) {
+	$destnets = $2;
+	$interfacelist = $1;
+    } elsif ( $interfacelist =~ /^([^:]+):$/ ) {
+	$add_snat_aliases = 0;
+	$interfacelist = $1;
+    } elsif ( $interfacelist =~ /^([^:]+):([^:]*)$/ ) {
+	my ( $one, $two ) = ( $1, $2 );
+	if ( $2 =~ /\./ ) {
+	    $interfacelist = $one;
+	    $destnets = $two;
+	}
+    }
+    #
+    # If there is no source or destination then allow all addresses
+    #
+    $networks = ALLIPv4 if $networks eq '-';
+    $destnets = ALLIPv4 if $destnets eq '-';
 
     #
     # Handle IPSEC options, if any
@@ -129,157 +159,153 @@ sub setup_one_masq($$$$$$$)
 	fatal_error "Non-empty IPSEC column requires policy match support in your kernel and iptables"  unless $globals{ORIGINAL_POLICY_MATCH};
 
 	if ( $ipsec =~ /^yes$/i ) {
-	    $rule .= '-m policy --pol ipsec --dir out ';
+	    $baserule .= '-m policy --pol ipsec --dir out ';
 	} elsif ( $ipsec =~ /^no$/i ) {
-	    $rule .= '-m policy --pol none --dir out ';
+	    $baserule .= '-m policy --pol none --dir out ';
 	} else {
-	    $rule .= do_ipsec_options $ipsec;
+	    $baserule .= do_ipsec_options $ipsec;
 	}
     } elsif ( $capabilities{POLICY_MATCH} ) {
-	$rule .= '-m policy --pol none --dir out ';
+	$baserule .= '-m policy --pol none --dir out ';
     }
 
-    #
-    # Leading '+'
-    #
-    $pre_nat = 1 if $fullinterface =~ s/^\+//;
-    #
-    # Parse the remaining part of the INTERFACE column
-    #
-    if ( $fullinterface =~ /^([^:]+)::([^:]*)$/ ) {
-	$add_snat_aliases = 0;
-	$destnets = $2;
-	$fullinterface = $1;
-    } elsif ( $fullinterface =~ /^([^:]+:[^:]+):([^:]+)$/ ) {
-	$destnets = $2;
-	$fullinterface = $1;
-    } elsif ( $fullinterface =~ /^([^:]+):$/ ) {
-	$add_snat_aliases = 0;
-	$fullinterface = $1;
-    } elsif ( $fullinterface =~ /^([^:]+):([^:]*)$/ ) {
-	my ( $one, $two ) = ( $1, $2 );
-	if ( $2 =~ /\./ ) {
-	    $fullinterface = $one;
-	    $destnets = $two;
-	}
-    }
-
-    #
-    # Isolate and verify the interface part
-    #
-    ( my $interface = $fullinterface ) =~ s/:.*//;
-
-    fatal_error "Unknown interface ($interface)" unless find_interface( $interface )->{root};
-
-    my $chainref = ensure_chain('nat', $pre_nat ? snat_chain $interface : masq_chain $interface);
-    #
-    # If there is no source or destination then allow all addresses
-    #
-    $networks = ALLIPv4 if $networks eq '-';
-    $destnets = ALLIPv4 if $destnets eq '-';
     #
     # Handle Protocol and Ports
     #
-    $rule .= do_proto $proto, $ports, '';
+    $baserule .= do_proto $proto, $ports, '';
+
     #
     # Handle Mark
     #
-    $rule .= do_test( $mark, 0xFF) if $mark ne '-';
+    $baserule .= do_test( $mark, 0xFF) if $mark ne '-';
 
-    my $detectaddress = 0;
-    my $exceptionrule = '';
-    my $randomize     = '';
-    #
-    # Parse the ADDRESSES column
-    #
-    if ( $addresses ne '-' ) {
-	if ( $addresses eq 'random' ) {
-	    $randomize = '--random ';
-	} else {
-	    $addresses =~ s/:random$// and $randomize = '--random ';
+    for my $fullinterface (split_list $interfacelist, 'interface' ) {
+	my $rule = '';
+	my $target = '-j MASQUERADE ';
+	#
+	# Isolate and verify the interface part
+	#
+	( my $interface = $fullinterface ) =~ s/:.*//;
 
-	    if ( $addresses =~ /^SAME:nodst:/ ) {
-		fatal_error "':random' is not supported by the SAME target" if $randomize;
-		$target = '-j SAME --nodst ';
-		$addresses =~ s/.*://;
-		for my $addr ( split /,/, $addresses ) {
-		    $target .= "--to $addr ";
-		}
-	    } elsif ( $addresses =~ /^SAME:/ ) {
-		fatal_error "':random' is not supported by the SAME target" if $randomize;
-		$target = '-j SAME ';
-		$addresses =~ s/.*://;
-		for my $addr ( split /,/, $addresses ) {
-		    $target .= "--to $addr ";
-		}
-	    } elsif ( $addresses eq 'detect' ) {
-		my $variable = get_interface_address $interface;
-		$target = "-j SNAT --to-source $variable";
-		
-		if ( interface_is_optional $interface ) {
-		    add_commands( $chainref,
-				  '',
-				  "if [ \"$variable\" != 0.0.0.0 ]; then" );
-		    incr_cmd_level( $chainref );
-		    $detectaddress = 1;
-		}
-	    } else {
-		my $addrlist = '';
-		for my $addr ( split /,/, $addresses ) {
-		    if ( $addr =~ /^.*\..*\..*\./ ) {
-			$target = '-j SNAT ';
-			$addrlist .= "--to-source $addr ";
-			$exceptionrule = do_proto( $proto, '', '' ) if $addr =~ /:/;
-		    } else {
-			$addr =~ s/^://;
-			$addrlist .= "--to-ports $addr ";
-			$exceptionrule = do_proto( $proto, '', '' );
-		    }
-		}
-		
-		$target .= $addrlist;
-	    }
+	if ( $interface =~ /(.*)[(](\w*)[)]$/ ) {
+	    $interface = $1;
+	    my $provider  = $2;
+	    $fullinterface =~ s/[(]\w*[)]//;
+	    my $realm = lookup_provider( $provider ) unless $provider =~ /^\d+$/;
+
+	    fatal_error "$provider is not a shared-interface provider" unless $realm;
+
+	    $rule .= "-m realm --realm $realm ";
 	}
 
-	$target .= $randomize;
-    } else {
-	$add_snat_aliases = 0;
-    }
-    #
-    # And Generate the Rule(s)
-    #
-    expand_rule( $chainref ,
-		 POSTROUTE_RESTRICT ,
-		 $rule ,
-		 $networks ,
-		 $destnets ,
-		 '' ,
-		 $target ,
-		 '' ,
-		 '' ,
-		 $exceptionrule );
+	fatal_error "Unknown interface ($interface)" unless my $interfaceref = known_interface( $interface );
 
-    if ( $detectaddress ) {
-	decr_cmd_level( $chainref );
-	add_command( $chainref , 'fi' );
-    }
+	unless ( $interfaceref->{root} ) {
+	    $rule .= "-o $interface ";
+	    $interface = $interfaceref->{name};
+	}
 
-    if ( $add_snat_aliases ) {
-	my ( $interface, $alias , $remainder ) = split( /:/, $fullinterface, 3 );
-	fatal_error "Invalid alias ($alias:$remainder)" if defined $remainder;
-	for my $address ( split /,/, $addresses ) {
-	    my ( $addrs, $port ) = split /:/, $address;
-	    next unless $addrs;
-	    next if $addrs eq 'detect';
-	    for my $addr ( ip_range_explicit $addrs ) {
-		unless ( $addresses_to_add{$addr} ) {
-		    emit "del_ip_addr $addr $interface" unless $config{RETAIN_ALIASES};
-		    $addresses_to_add{$addr} = 1;
-		    if ( defined $alias ) {
-			push @addresses_to_add, $addr, "$interface:$alias";
-			$alias++;
-		    } else {
-			push @addresses_to_add, $addr, $interface;
+	my $chainref = ensure_chain('nat', $pre_nat ? snat_chain $interface : masq_chain $interface);
+
+	my $detectaddress = 0;
+	my $exceptionrule = '';
+	my $randomize     = '';
+	#
+	# Parse the ADDRESSES column
+	#
+	if ( $addresses ne '-' ) {
+	    if ( $addresses eq 'random' ) {
+		$randomize = '--random ';
+	    } else {
+		$addresses =~ s/:random$// and $randomize = '--random ';
+
+		if ( $addresses =~ /^SAME:nodst:/ ) {
+		    fatal_error "':random' is not supported by the SAME target" if $randomize;
+		    $target = '-j SAME --nodst ';
+		    $addresses =~ s/.*://;
+		    for my $addr ( split_list $addresses, 'address' ) {
+			$target .= "--to $addr ";
+		    }
+		} elsif ( $addresses =~ /^SAME:/ ) {
+		    fatal_error "':random' is not supported by the SAME target" if $randomize;
+		    $target = '-j SAME ';
+		    $addresses =~ s/.*://;
+		    for my $addr ( split_list $addresses, 'address' ) {
+			$target .= "--to $addr ";
+		    }
+		} elsif ( $addresses eq 'detect' ) {
+		    my $variable = get_interface_address $interface;
+		    $target = "-j SNAT --to-source $variable";
+
+		    if ( interface_is_optional $interface ) {
+			add_commands( $chainref,
+				      '',
+				      "if [ \"$variable\" != 0.0.0.0 ]; then" );
+			incr_cmd_level( $chainref );
+			$detectaddress = 1;
+		    }
+		} elsif ( $addresses eq 'NONAT' ) {
+		    $target = '-j RETURN';
+		    $add_snat_aliases = 0;
+		} else {
+		    my $addrlist = '';
+		    for my $addr ( split_list $addresses , 'address' ) {
+			if ( $addr =~ /^.*\..*\..*\./ ) {
+			    $target = '-j SNAT ';
+			    $addrlist .= "--to-source $addr ";
+			    $exceptionrule = do_proto( $proto, '', '' ) if $addr =~ /:/;
+			} else {
+			    $addr =~ s/^://;
+			    $addrlist .= "--to-ports $addr ";
+			    $exceptionrule = do_proto( $proto, '', '' );
+			}
+		    }
+
+		    $target .= $addrlist;
+		}
+	    }
+
+	    $target .= $randomize;
+	} else {
+	    $add_snat_aliases = 0;
+	}
+	#
+	# And Generate the Rule(s)
+	#
+	expand_rule( $chainref ,
+		     POSTROUTE_RESTRICT ,
+		     $baserule . $rule ,
+		     $networks ,
+		     $destnets ,
+		     '' ,
+		     '' ,
+		     $target ,
+		     '' ,
+		     '' ,
+		     $exceptionrule );
+
+	if ( $detectaddress ) {
+	    decr_cmd_level( $chainref );
+	    add_command( $chainref , 'fi' );
+	}
+
+	if ( $add_snat_aliases ) {
+	    my ( $interface, $alias , $remainder ) = split( /:/, $fullinterface, 3 );
+	    fatal_error "Invalid alias ($alias:$remainder)" if defined $remainder;
+	    for my $address ( split_list $addresses, 'address' ) {
+		my ( $addrs, $port ) = split /:/, $address;
+		next unless $addrs;
+		next if $addrs eq 'detect';
+		for my $addr ( ip_range_explicit $addrs ) {
+		    unless ( $addresses_to_add{$addr} ) {
+			emit "del_ip_addr $addr $interface" unless $config{RETAIN_ALIASES};
+			$addresses_to_add{$addr} = 1;
+			if ( defined $alias ) {
+			    push @addresses_to_add, $addr, "$interface:$alias";
+			    $alias++;
+			} else {
+			    push @addresses_to_add, $addr, $interface;
+			}
 		    }
 		}
 	    }
@@ -298,7 +324,7 @@ sub setup_masq()
     my $fn = open_file 'masq';
 
     first_entry( sub { progress_message2 "$doing $fn..."; require_capability 'NAT_ENABLED' , 'a non-empty masq file' , 's'; } );
-    
+
     while ( read_a_line ) {
 
 	my ($fullinterface, $networks, $addresses, $proto, $ports, $ipsec, $mark ) = split_line1 2, 7, 'masq file';
@@ -353,6 +379,16 @@ sub do_one_nat( $$$$$ )
 
     my $policyin = '';
     my $policyout = '';
+    my $rulein = '';
+    my $ruleout = '';
+
+    fatal_error "Unknown interface ($interface)" unless my $interfaceref = known_interface( $interface );
+
+    unless ( $interfaceref->{root} ) {
+	$rulein  = "-i $interface ";
+	$ruleout = "-o $interface ";
+	$interface = $interfaceref->{name};
+    }
 
     if ( $capabilities{POLICY_MATCH} ) {
 	$policyin = ' -m policy --pol none --dir in';
@@ -362,11 +398,7 @@ sub do_one_nat( $$$$$ )
     fatal_error "Invalid nat file entry" unless defined $interface && defined $internal;
 
     if ( $add_ip_aliases ) {
-	if ( defined( $alias ) && $alias eq '' ) {
-	    $add_ip_aliases = '';
-	} else {
-	    emit "del_ip_addr $external $interface" unless $config{RETAIN_ALIASES};
-	}
+	$add_ip_aliases = '' if defined( $alias ) && $alias eq '';
     }
 
     validate_nat_column 'ALL INTERFACES', \$allints;
@@ -376,8 +408,8 @@ sub do_one_nat( $$$$$ )
 	add_nat_rule 'nat_in' ,  "-d $external $policyin  -j DNAT --to-destination $internal";
 	add_nat_rule 'nat_out' , "-s $internal $policyout -j SNAT --to-source $external";
     } else {
-	add_nat_rule input_chain( $interface ) ,  "-d $external $policyin -j DNAT --to-destination $internal";
-	add_nat_rule output_chain( $interface ) , "-s $internal $policyout -j SNAT --to-source $external";
+	add_nat_rule input_chain( $interface ) ,  $rulein  . "-d $external $policyin -j DNAT --to-destination $internal";
+	add_nat_rule output_chain( $interface ) , $ruleout . "-s $internal $policyout -j SNAT --to-source $external";
     }
 
     add_nat_rule 'OUTPUT' , "-d $external $policyout -j DNAT --to-destination $internal " if $localnat;
@@ -389,7 +421,6 @@ sub do_one_nat( $$$$$ )
 	}
     }
 
-    progress_message "   NAT entry \"$currentline\" $done";
 }
 
 #
@@ -400,15 +431,24 @@ sub setup_nat() {
     my $fn = open_file 'nat';
 
     first_entry( sub { progress_message2 "$doing $fn..."; require_capability 'NAT_ENABLED' , 'a non-empty nat file' , 's'; } );
-    
+
     while ( read_a_line ) {
 
-	my ( $external, $interface, $internal, $allints, $localnat ) = split_line1 3, 5, 'nat file';
+	my ( $external, $interfacelist, $internal, $allints, $localnat ) = split_line1 3, 5, 'nat file';
 
 	if ( $external eq 'COMMENT' ) {
 	    process_comment;
 	} else {
-	    do_one_nat $external, $interface, $internal, $allints, $localnat;
+	    ( $interfacelist, my $digit ) = split /:/, $interfacelist;
+
+	    $digit = defined $digit ? ":$digit" : '';
+
+	    for my $interface ( split_list $interfacelist , 'interface' ) {
+		fatal_error "Invalid Interface List ($interfacelist)" unless defined $interface && $interface ne '';
+		do_one_nat $external, "${interface}${digit}", $internal, $allints, $localnat;
+	    }
+
+	    progress_message "   NAT entry \"$currentline\" $done";
 	}
 
     }
@@ -427,20 +467,32 @@ sub setup_netmap() {
 
     while ( read_a_line ) {
 
-	my ( $type, $net1, $interface, $net2 ) = split_line 4, 4, 'netmap file';
+	my ( $type, $net1, $interfacelist, $net2 ) = split_line 4, 4, 'netmap file';
 
-	fatal_error "Unknown Interface ($interface)" unless known_interface $interface;
+	for my $interface ( split_list $interfacelist, 'interface' ) {
 
-	if ( $type eq 'DNAT' ) {
-	    add_rule ensure_chain( 'nat' , input_chain $interface )  , "-d $net1 -j NETMAP --to $net2";
-	} elsif ( $type eq 'SNAT' ) {
-	    add_rule ensure_chain( 'nat' , output_chain $interface ) , "-s $net1 -j NETMAP --to $net2";
-	} else {
-	    fatal_error "Invalid type ($type)";
+	    my $rulein = '';
+	    my $ruleout = '';
+	    my $iface = $interface;
+
+	    fatal_error "Unknown interface ($interface)" unless my $interfaceref = find_interface( $interface );
+
+	    unless ( $interfaceref->{root} ) {
+		$rulein  = "-i $interface ";
+		$ruleout = "-o $interface ";
+		$interface = $interfaceref->{name};
+	    }
+
+	    if ( $type eq 'DNAT' ) {
+		add_rule ensure_chain( 'nat' , input_chain $interface )  , $rulein  . "-d $net1 -j NETMAP --to $net2";
+	    } elsif ( $type eq 'SNAT' ) {
+		add_rule ensure_chain( 'nat' , output_chain $interface ) , $ruleout . "-s $net1 -j NETMAP --to $net2";
+	    } else {
+		fatal_error "Invalid type ($type)";
+	    }
+
+	    progress_message "   Network $net1 on $iface mapped to $net2 ($type)";
 	}
-
-	progress_message "   Network $net1 on $interface mapped to $net2 ($type)";
-
     }
 
 }
@@ -448,11 +500,17 @@ sub setup_netmap() {
 sub add_addresses () {
     if ( @addresses_to_add ) {
 	my $arg = '';
+	my $addresses = 0;
 
 	while ( @addresses_to_add ) {
 	    my $addr      = shift @addresses_to_add;
 	    my $interface = shift @addresses_to_add;
 	    $arg = "$arg $addr $interface";
+	    unless ( $config{RETAIN_ALIASES} ) {
+		emit '' unless $addresses++;
+		$interface =~ s/:.*//;
+		emit "del_ip_addr $addr $interface";
+	    }
 	}
 
 	emit "add_ip_aliases $arg";
