@@ -1,9 +1,9 @@
 #
-# Shorewall-perl 4.0 -- /usr/share/shorewall-perl/Shorewall/Policy.pm
+# Shorewall-perl 4.2 -- /usr/share/shorewall-perl/Shorewall/Policy.pm
 #
 #     This program is under GPL [http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt]
 #
-#     (c) 2007 - Tom Eastep (teastep@shorewall.net)
+#     (c) 2007,2008 - Tom Eastep (teastep@shorewall.net)
 #
 #       Complete documentation is available at http://shorewall.net
 #
@@ -32,9 +32,9 @@ use Shorewall::Actions;
 use strict;
 
 our @ISA = qw(Exporter);
-our @EXPORT = qw( validate_policy apply_policy_rules complete_standard_chain sub setup_syn_flood_chains );
+our @EXPORT = qw( validate_policy apply_policy_rules complete_standard_chain setup_syn_flood_chains );
 our @EXPORT_OK = qw(  );
-our $VERSION = 4.0.6;
+our $VERSION = 4.2.4;
 
 # @policy_chains is a list of references to policy chains in the filter table
 
@@ -131,7 +131,7 @@ sub add_or_modify_policy_chain( $$ ) {
     my ( $zone, $zone1 ) = @_;
     my $chain    = "${zone}2${zone1}";
     my $chainref = $filter_table->{$chain};
-    
+
     if ( $chainref ) {
 	unless( $chainref->{is_policy} ) {
 	    convert_to_policy_chain( $chainref, $zone, $zone1, 'CONTINUE', OPTIONAL );
@@ -149,7 +149,7 @@ sub print_policy($$$$) {
 	    my ( $sourceref, $destref ) = ( find_zone($source) ,find_zone( $dest ) );
 	    warning_message "CONTINUE policy between two un-nested zones ($source, $dest)" if ! ( @{$sourceref->{parents}} || @{$destref->{parents}} );
 	}
-	progress_message "   Policy for $source to $dest is $policy using chain $chain" unless $source eq $dest;
+	progress_message_nocompress "   Policy for $source to $dest is $policy using chain $chain" unless $source eq $dest;
     }
 }
 
@@ -172,6 +172,7 @@ sub validate_policy()
 	        NFQUEUE_DEFAULT => 'NFQUEUE' );
 
     my $zone;
+    my @zonelist = $config{EXPAND_POLICIES} ? all_zones : ( all_zones, 'all' );
 
     for my $option qw/DROP_DEFAULT REJECT_DEFAULT ACCEPT_DEFAULT QUEUE_DEFAULT NFQUEUE_DEFAULT/ {
 	my $action = $config{$option};
@@ -211,18 +212,19 @@ sub validate_policy()
 
     while ( read_a_line ) {
 
-	my ( $client, $server, $originalpolicy, $loglevel, $synparams ) = split_line 3, 5, 'policy file';
+	my ( $client, $server, $originalpolicy, $loglevel, $synparams, $connlimit ) = split_line 3, 6, 'policy file';
 
 	$loglevel  = '' if $loglevel  eq '-';
 	$synparams = '' if $synparams eq '-';
+	$connlimit = '' if $connlimit eq '-';
 
 	my $clientwild = ( "\L$client" eq 'all' );
 
-	fatal_error "Undefined zone $client" unless $clientwild || defined_zone( $client );
+	fatal_error "Undefined zone ($client)" unless $clientwild || defined_zone( $client );
 
 	my $serverwild = ( "\L$server" eq 'all' );
 
-	fatal_error "Undefined zone $server" unless $serverwild || defined_zone( $server );
+	fatal_error "Undefined zone ($server)" unless $serverwild || defined_zone( $server );
 
 	my ( $policy, $default, $remainder ) = split( /:/, $originalpolicy, 3 );
 
@@ -230,7 +232,7 @@ sub validate_policy()
 
 	fatal_error "Invalid default action ($default:$remainder)" if defined $remainder;
 
-	( $policy , my $queue ) = split( '/' , $policy );
+	( $policy , my $queue ) = get_target_param $policy;
 
 	if ( $default ) {
 	    if ( "\L$default" eq 'none' ) {
@@ -251,14 +253,14 @@ sub validate_policy()
 	    $default = $default_actions{$policy} || '';
 	}
 
-	fatal_error "Invalid policy $policy" unless exists $validpolicies{$policy};
+	fatal_error "Invalid policy ($policy)" unless exists $validpolicies{$policy};
 
 	if ( defined $queue ) {
-	    fatal_error "Invalid policy ($policy/$queue)" unless $policy eq 'NFQUEUE';
+	    fatal_error "Invalid policy ($policy($queue))" unless $policy eq 'NFQUEUE';
 	    require_capability( 'NFQUEUE_TARGET', 'An NFQUEUE Policy', 's' ); 
-	    $queue = numeric_value( $queue );
-	    fatal_error "Invalid NFQUEUE queue number ($queue)" if $queue > 65535;
-	    $policy = "$policy/$queue";
+	    my $queuenum = numeric_value( $queue );
+	    fatal_error "Invalid NFQUEUE queue number ($queue)" unless defined( $queuenum) && $queuenum <= 65535;
+	    $policy = "NFQUEUE --queue-num $queuenum";
 	} elsif ( $policy eq 'NONE' ) {
 	    fatal_error "NONE policy not allowed with \"all\""
 		if $clientwild || $serverwild;
@@ -267,7 +269,7 @@ sub validate_policy()
 	}
 
 	unless ( $clientwild || $serverwild ) {
-	    if ( zone_type( $server ) eq 'bport4' ) {
+	    if ( zone_type( $server ) eq 'bport' ) {
 		fatal_error "Invalid policy - DEST zone is a Bridge Port zone but the SOURCE zone is not associated with the same bridge"
 		    unless find_zone( $client )->{bridge} eq find_zone( $server)->{bridge} || single_interface( $client ) eq find_zone( $server )->{bridge};
 	    }
@@ -299,8 +301,12 @@ sub validate_policy()
 
 	$chainref->{loglevel}  = validate_level( $loglevel ) if defined $loglevel && $loglevel ne '';
 
-	if ( $synparams ne '' ) {
-	    $chainref->{synparams} = do_ratelimit $synparams, 'ACCEPT';
+	if ( $synparams ne '' || $connlimit ne '' ) {
+	    my $value = '';
+	    fatal_error "Invalid CONNLIMIT ($connlimit)" if $connlimit =~ /^!/;
+	    $value  = do_ratelimit $synparams, 'ACCEPT'  if $synparams ne '';
+	    $value .= do_connlimit $connlimit            if $connlimit ne '';
+	    $chainref->{synparams} = $value;
 	    $chainref->{synchain}  = $chain
 	}
 
@@ -308,8 +314,8 @@ sub validate_policy()
 
 	if ( $clientwild ) {
 	    if ( $serverwild ) {
-		for my $zone ( all_zones , 'all' ) {
-		    for my $zone1 ( all_zones , 'all' ) {
+		for my $zone ( @zonelist ) {
+		    for my $zone1 ( @zonelist ) {
 			set_policy_chain $client, $server, "${zone}2${zone1}", $chainref, $policy;
 			print_policy $zone, $zone1, $policy, $chain;
 		    }
@@ -321,7 +327,7 @@ sub validate_policy()
 		}
 	    }
 	} elsif ( $serverwild ) {
-	    for my $zone ( all_zones , 'all' ) {
+	    for my $zone ( @zonelist ) {
 		set_policy_chain $client, $server, "${client}2${zone}", $chainref, $policy;
 		print_policy $client, $zone, $policy, $chain;
 	    }
@@ -349,19 +355,13 @@ sub policy_rules( $$$$$ ) {
 	add_rule $chainref, "-j $default" if $default && $default ne 'none';
 	log_rule $loglevel , $chainref , $target , '' if $loglevel ne '';
 	fatal_error "Null target in policy_rules()" unless $target;
-	if ( $target eq 'REJECT' ) {
-	    $target = 'reject';
-	} elsif ( $target =~ /^NFQUEUE/ ) {
-	    my $queue = ( split( '/', $target) )[1] || 0;
-	    $target = "NFQUEUE --queue-num $queue";
-	}
 
-	add_rule( $chainref , "-j $target" ) unless $target eq 'CONTINUE';
+	add_jump( $chainref , $target eq 'REJECT' ? 'reject' : $target, 1 ) unless $target eq 'CONTINUE';
     }
 }
 
 sub report_syn_flood_protection() {
-    progress_message '      Enabled SYN flood protection';
+    progress_message_nocompress '      Enabled SYN flood protection';
 }
 
 sub default_policy( $$$ ) {
@@ -382,7 +382,7 @@ sub default_policy( $$$ ) {
 		report_syn_flood_protection;
 		policy_rules $chainref , $policy , $loglevel , $default, $config{MULTICAST};
 	    } else {
-		add_rule $chainref,  "-j $policyref->{name}";
+		add_jump $chainref,  $policyref, 1;
 		$chainref = $policyref;
 	    }
 	} elsif ( $policy eq 'CONTINUE' ) {
@@ -390,12 +390,12 @@ sub default_policy( $$$ ) {
 	    policy_rules $chainref , $policy , $loglevel , $default, $config{MULTICAST};
 	} else {
 	    report_syn_flood_protection if $synparams;
-	    add_rule $chainref , "-j $policyref->{name}";
+	    add_jump $chainref , $policyref, 1;
 	    $chainref = $policyref;
 	}
     }
 
-    progress_message "   Policy $policy from $_[1] to $_[2] using chain $chainref->{name}";
+    progress_message_nocompress "   Policy $policy from $_[1] to $_[2] using chain $chainref->{name}";
 
 }
 
@@ -418,7 +418,6 @@ sub apply_policy_rules() {
 		run_user_exit $chainref;
 		policy_rules $chainref , $policy, $loglevel , $default, $config{MULTICAST};
 	    }
-
 	}
     }
 
@@ -443,22 +442,22 @@ sub apply_policy_rules() {
 #	- If no applicable policy is found, add rules for an assummed
 #	  policy of DROP INFO
 #
-sub complete_standard_chain ( $$$ ) {
-    my ( $stdchainref, $zone, $zone2 ) = @_;
+sub complete_standard_chain ( $$$$ ) {
+    my ( $stdchainref, $zone, $zone2, $default ) = @_;
 
     add_rule $stdchainref, '-m state --state ESTABLISHED,RELATED -j ACCEPT' unless $config{FASTACCEPT};
 
     run_user_exit $stdchainref;
 
     my $ruleschainref = $filter_table->{"${zone}2${zone2}"};
-    my ( $policy, $loglevel, $default ) = ( 'DROP', 6, $config{DROP_DEFAULT} );
+    my ( $policy, $loglevel, $defaultaction ) = ( $default , 6, $config{$default . '_DEFAULT'} );
     my $policychainref;
 
     $policychainref = $filter_table->{$ruleschainref->{policychain}} if $ruleschainref;
 
-    ( $policy, $loglevel, $default ) = @{$policychainref}{'policy', 'loglevel', 'default' } if $policychainref;
+    ( $policy, $loglevel, $defaultaction ) = @{$policychainref}{'policy', 'loglevel', 'default' } if $policychainref;
 
-    policy_rules $stdchainref , $policy , $loglevel, $default, 0;
+    policy_rules $stdchainref , $policy , $loglevel, $defaultaction, 0;
 }
 
 #
