@@ -1,5 +1,5 @@
 #
-# Shorewall 4.0 -- /usr/share/shorewall-perl/Shorewall/Proc.pm
+# Shorewall 4.2 -- /usr/share/shorewall-perl/Shorewall/Proc.pm
 #
 #     This program is under GPL [http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt]
 #
@@ -41,7 +41,7 @@ our @EXPORT = qw(
 		 setup_forwarding
 		 );
 our @EXPORT_OK = qw( );
-our $VERSION = 4.0.6;
+our $VERSION = 4.2.4;
 
 #
 # ARP Filtering
@@ -123,7 +123,7 @@ sub setup_route_filtering() {
 	    emit 'echo 0 > /proc/sys/net/ipv4/conf/default/rp_filter';
 	}
 
-	emit "[ -n \"\$NOROUTES\" ] || ip route flush cache";
+	emit "[ -n \"\$NOROUTES\" ] || ip -4 route flush cache";
     }
 }
 
@@ -145,7 +145,10 @@ sub setup_martian_logging() {
 
 	    emit ( 'for file in /proc/sys/net/ipv4/conf/*; do',
 		   "    [ -f \$file/log_martians ] && echo $val > \$file/log_martians",
-		   'done' );
+		   'done',
+		   '' );
+
+	    emit 'echo 0 > /proc/sys/net/ipv4/conf/all/log_martians' if $val == 1;
 	}
 
 	for my $interface ( @$interfaces ) {
@@ -159,21 +162,14 @@ sub setup_martian_logging() {
 		   "    error_message \"WARNING: Cannot set Martian logging on $interface\"") unless interface_is_optional( $interface);
 	    emit   "fi\n";
 	}
-
-	if ( $config{LOG_MARTIANS} eq 'on' ) {
-	    emit 'echo 1 > /proc/sys/net/ipv4/conf/all/log_martians';
-	    emit 'echo 1 > /proc/sys/net/ipv4/conf/default/log_martians';
-	} elsif ( $config{LOG_MARTIANS} eq 'off' ) {
-	    emit 'echo 0 > /proc/sys/net/ipv4/conf/all/log_martians';
-	    emit 'echo 0 > /proc/sys/net/ipv4/conf/default/log_martians';
-	}
     }
 }
 
 #
 # Source Routing
 #
-sub setup_source_routing() {
+sub setup_source_routing( $ ) {
+    my $family = shift;
 
     save_progress_message 'Setting up Accept Source Routing...';
 
@@ -185,7 +181,7 @@ sub setup_source_routing() {
 	save_progress_message 'Setting up Source Routing...';
 
 	for my $interface ( @$interfaces ) {
-	    my $file = "/proc/sys/net/ipv4/conf/$interface/accept_source_route";
+	    my $file = "/proc/sys/net/ipv$family/conf/$interface/accept_source_route";
 	    my $value = get_interface_option $interface, 'sourceroute';
 
 	    emit ( "if [ -f $file ]; then" ,
@@ -197,16 +193,55 @@ sub setup_source_routing() {
     }
 }
 
-sub setup_forwarding() {
-    if ( $config{IP_FORWARDING} eq 'on' ) {
-	emit '        echo 1 > /proc/sys/net/ipv4/ip_forward';
-	emit '        progress_message2 IP Forwarding Enabled';
-    } elsif ( $config{IP_FORWARDING} eq 'off' ) {
-	emit '        echo 0 > /proc/sys/net/ipv4/ip_forward';
-	emit '        progress_message2 IP Forwarding Disabled!';
-    }
+sub setup_forwarding( $$ ) {
+    my ( $family, $first ) = @_;
 
-    emit '';
+    if ( $family == F_IPV4 ) {
+	if ( $config{IP_FORWARDING} eq 'on' ) {
+	    emit '        echo 1 > /proc/sys/net/ipv4/ip_forward';
+	    emit '        progress_message2 IPv4 Forwarding Enabled';
+	} elsif ( $config{IP_FORWARDING} eq 'off' ) {
+	    emit '        echo 0 > /proc/sys/net/ipv4/ip_forward';
+	    emit '        progress_message2 IPv4 Forwarding Disabled!';
+	}
+
+	emit '';
+    } else {
+	if ( $config{IP_FORWARDING} eq 'on' ) {
+	    emit '        echo 1 > /proc/sys/net/ipv6/conf/all/forwarding';
+	    emit '        progress_message2 IPv6 Forwarding Enabled';
+	} elsif ( $config{IP_FORWARDING} eq 'off' ) {
+	    emit '        echo 0 > /proc/sys/net/ipv6/conf/all/forwarding';
+	    emit '        progress_message2 IPv6 Forwarding Disabled!';
+	}
+
+	emit '';
+
+	my $interfaces = find_interfaces_by_option 'forward';
+
+	if ( @$interfaces ) {
+	    progress_message2 "$doing Interface forwarding..." if $first;
+
+	    push_indent;
+	    push_indent;
+
+	    save_progress_message 'Setting up IPv6 Interface Forwarding...';
+
+	    for my $interface ( @$interfaces ) {
+		my $file = "/proc/sys/net/ipv6/conf/$interface/forwarding";
+		my $value = get_interface_option $interface, 'forward';
+
+		emit ( "if [ -f $file ]; then" ,
+		       "    echo $value > $file" );
+		emit ( 'else' ,
+		       "    error_message \"WARNING: Cannot set IPv6 forwarding on $interface\"" ) unless interface_is_optional( $interface);
+		emit   "fi\n";
+	    }
+
+	    pop_indent;
+	    pop_indent;
+	}
+    }
 }
 
 1;
