@@ -1,5 +1,5 @@
 #
-# Shorewall-perl 4.0 -- /usr/share/shorewall-perl/Shorewall/Tunnels.pm
+# Shorewall-perl 4.2 -- /usr/share/shorewall-perl/Shorewall/Tunnels.pm
 #
 #     This program is under GPL [http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt]
 #
@@ -34,7 +34,7 @@ use strict;
 our @ISA = qw(Exporter);
 our @EXPORT = qw( setup_tunnels );
 our @EXPORT_OK = ( );
-our $VERSION = 4.0.6;
+our $VERSION = 4.2.4;
 
 #
 # Here starts the tunnel stuff -- we really should get rid of this crap...
@@ -50,7 +50,7 @@ sub setup_tunnels() {
 
 	my $noah = 1;
 
-	fatal_error "Invalid IPSEC modified ($qualifier:$remainder)" if defined $remainder;
+	fatal_error "Invalid IPSEC modifier ($qualifier:$remainder)" if defined $remainder;
 
 	if ( defined $qualifier ) {
 	    if ( $qualifier eq 'ah' ) {
@@ -61,48 +61,47 @@ sub setup_tunnels() {
 	    }
 	}
 
-	my $options = '-m state --state NEW -j ACCEPT';
+	my $options = $globals{UNTRACKED} ? '-m state --state NEW,UNTRACKED -j ACCEPT' : '-m state --state NEW -j ACCEPT';
 
-	add_rule $inchainref,  "-p 50 $source -j ACCEPT";
-	add_rule $outchainref, "-p 50 $dest   -j ACCEPT";
+	add_tunnel_rule $inchainref,  "-p 50 $source -j ACCEPT";
+	add_tunnel_rule $outchainref, "-p 50 $dest   -j ACCEPT";
 
 	unless ( $noah ) {
-	    add_rule $inchainref,  "-p 51 $source -j ACCEPT";
-	    add_rule $outchainref, "-p 51 $dest   -j ACCEPT";
+	    add_tunnel_rule $inchainref,  "-p 51 $source -j ACCEPT";
+	    add_tunnel_rule $outchainref, "-p 51 $dest   -j ACCEPT";
 	}
 
-	add_rule $outchainref,  "-p udp $dest --dport 500 $options";
-
 	if ( $kind eq 'ipsec' ) {
-	    add_rule $inchainref, "-p udp $source --dport 500 $options";
+	    add_tunnel_rule $inchainref,  "-p udp $source --dport 500 $options";
+	    add_tunnel_rule $outchainref, "-p udp $dest   --dport 500 $options";
 	} else {
-	    add_rule $inchainref,  "-p udp $source -m multiport --dports 500,4500 $options";
-	    add_rule $outchainref, "-p udp $dest   -m multiport --dports 500,4500 $options";
+	    add_tunnel_rule $inchainref,  "-p udp $source -m multiport --dports 500,4500 $options";
+	    add_tunnel_rule $outchainref, "-p udp $dest   -m multiport --dports 500,4500 $options";
 	}
 
 	unless ( $gatewayzones eq '-' ) {
-	    for my $zone ( split /,/, $gatewayzones ) {
+	    for my $zone ( split_list $gatewayzones, 'zone' ) {
 		my $type = zone_type( $zone );
-		fatal_error "Invalid zone ($zone) for GATEWAY ZONE" if $type eq 'firewall' || $type eq 'bport4';
+		fatal_error "Invalid zone ($zone) for GATEWAY ZONE" if $type eq 'firewall' || $type eq 'bport';
 		$inchainref  = ensure_filter_chain "${zone}2${fw}", 1;
 		$outchainref = ensure_filter_chain "${fw}2${zone}", 1;
 
 		unless ( $capabilities{POLICY_MATCH} ) {
-		    add_rule $inchainref,  "-p 50 $source -j ACCEPT";
-		    add_rule $outchainref, "-p 50 $dest -j ACCEPT";
+		    add_tunnel_rule $inchainref,  "-p 50 $source -j ACCEPT";
+		    add_tunnel_rule $outchainref, "-p 50 $dest -j ACCEPT";
 
 		    unless ( $noah ) {
-			add_rule $inchainref,  "-p 51 $source -j ACCEPT";
-			add_rule $outchainref, "-p 51 $dest -j ACCEPT";
+			add_tunnel_rule $inchainref,  "-p 51 $source -j ACCEPT";
+			add_tunnel_rule $outchainref, "-p 51 $dest -j ACCEPT";
 		    }
 		}
 
 		if ( $kind eq 'ipsec' ) {
-		    add_rule $inchainref,  "-p udp $source --dport 500 $options";
-		    add_rule $outchainref, "-p udp $dest --dport 500 $options";
+		    add_tunnel_rule $inchainref,  "-p udp $source --dport 500 $options";
+		    add_tunnel_rule $outchainref, "-p udp $dest --dport 500 $options";
 		} else {
-		    add_rule $inchainref,  "-p udp $source -m multiport --dports 500,4500 $options";
-		    add_rule $outchainref, "-p udp $dest -m multiport --dports 500,4500 $options";
+		    add_tunnel_rule $inchainref,  "-p udp $source -m multiport --dports 500,4500 $options";
+		    add_tunnel_rule $outchainref, "-p udp $dest -m multiport --dports 500,4500 $options";
 		}
 	    }
 	}
@@ -111,24 +110,24 @@ sub setup_tunnels() {
     sub setup_one_other {
 	my ($inchainref, $outchainref, $source, $dest , $protocol) = @_;
 
-	add_rule $inchainref ,  "-p $protocol $source -j ACCEPT";
-	add_rule $outchainref , "-p $protocol $dest -j ACCEPT";
+	add_tunnel_rule $inchainref ,  "-p $protocol $source -j ACCEPT";
+	add_tunnel_rule $outchainref , "-p $protocol $dest -j ACCEPT";
     }
 
     sub setup_pptp_client {
 	my ($inchainref, $outchainref, $kind, $source, $dest ) = @_;
 
-	add_rule $outchainref,  "-p 47 $dest -j ACCEPT";
-	add_rule $inchainref,   "-p 47 $source -j ACCEPT";
-	add_rule $outchainref,  "-p tcp --dport 1723 $dest -j ACCEPT"
+	add_tunnel_rule $outchainref,  "-p 47 $dest -j ACCEPT";
+	add_tunnel_rule $inchainref,   "-p 47 $source -j ACCEPT";
+	add_tunnel_rule $outchainref,  "-p tcp --dport 1723 $dest -j ACCEPT"
 	}
 
     sub setup_pptp_server {
 	my ($inchainref, $outchainref, $kind, $source, $dest ) = @_;
 
-	add_rule $inchainref,  "-p 47 $dest -j ACCEPT";
-	add_rule $outchainref, "-p 47 $source -j ACCEPT";
-	add_rule $inchainref,  "-p tcp --dport 1723 $dest -j ACCEPT"
+	add_tunnel_rule $inchainref,  "-p 47 $dest -j ACCEPT";
+	add_tunnel_rule $outchainref, "-p 47 $source -j ACCEPT";
+	add_tunnel_rule $inchainref,  "-p tcp --dport 1723 $dest -j ACCEPT"
 	}
 
     sub setup_one_openvpn {
@@ -152,8 +151,8 @@ sub setup_tunnels() {
 	    }
 	}
 
-	add_rule $inchainref,  "-p $protocol $source --dport $port -j ACCEPT";
-	add_rule $outchainref, "-p $protocol $dest --dport $port -j ACCEPT";
+	add_tunnel_rule $inchainref,  "-p $protocol $source --dport $port -j ACCEPT";
+	add_tunnel_rule $outchainref, "-p $protocol $dest --dport $port -j ACCEPT";
     }
 
     sub setup_one_openvpn_client {
@@ -177,8 +176,8 @@ sub setup_tunnels() {
 	    }
 	}
 
-	add_rule $inchainref,  "-p $protocol $source --sport $port -j ACCEPT";
-	add_rule $outchainref, "-p $protocol $dest --dport $port -j ACCEPT";
+	add_tunnel_rule $inchainref,  "-p $protocol $source --sport $port -j ACCEPT";
+	add_tunnel_rule $outchainref, "-p $protocol $dest --dport $port -j ACCEPT";
     }
 
     sub setup_one_openvpn_server {
@@ -202,8 +201,8 @@ sub setup_tunnels() {
 	    }
 	}
 
-	add_rule $inchainref,  "-p $protocol $source --dport $port -j ACCEPT";
-	add_rule $outchainref, "-p $protocol $dest --sport $port -j ACCEPT";
+	add_tunnel_rule $inchainref,  "-p $protocol $source --dport $port -j ACCEPT";
+	add_tunnel_rule $outchainref, "-p $protocol $dest --sport $port -j ACCEPT";
     }
 
     sub setup_one_l2tp {
@@ -211,8 +210,8 @@ sub setup_tunnels() {
 
 	fatal_error "Unknown option ($1)" if $kind =~ /^.*?:(.*)$/;
 
-	add_rule $inchainref,  "-p udp $source --sport 1701 --dport 1701 -j ACCEPT";
-	add_rule $outchainref, "-p udp $dest   --sport 1701 --dport 1701 -j ACCEPT";
+	add_tunnel_rule $inchainref,  "-p udp $source --sport 1701 --dport 1701 -j ACCEPT";
+	add_tunnel_rule $outchainref, "-p udp $dest   --sport 1701 --dport 1701 -j ACCEPT";
     }
 
     sub setup_one_generic {
@@ -229,8 +228,8 @@ sub setup_tunnels() {
 	    ( $kind, $protocol ) = split /:/ , $kind if $kind =~ /.*:.*/;
 	}
 
-	add_rule $inchainref,  "-p $protocol $source $port -j ACCEPT";
-	add_rule $outchainref, "-p $protocol $dest $port -j ACCEPT";
+	add_tunnel_rule $inchainref,  "-p $protocol $source $port -j ACCEPT";
+	add_tunnel_rule $outchainref, "-p $protocol $dest $port -j ACCEPT";
     }
 
     sub setup_one_tunnel($$$$) {
@@ -238,12 +237,12 @@ sub setup_tunnels() {
 
 	my $zonetype = zone_type( $zone );
 
-	fatal_error "Invalid zone ($zone) for tunnel ZONE" if $zonetype eq 'firewall' || $zonetype eq 'bport4';
+	fatal_error "Invalid tunnel ZONE ($zone)" if $zonetype eq 'firewall' || $zonetype eq 'bport';
 
 	my $inchainref  = ensure_filter_chain "${zone}2${fw}", 1;
 	my $outchainref = ensure_filter_chain "${fw}2${zone}", 1;
 
-	$gateway = ALLIPv4 if $gateway eq '-';
+	$gateway = ALLIP if $gateway eq '-';
 
 	my $source = match_source_net $gateway;
 	my $dest   = match_dest_net   $gateway;
