@@ -1,5 +1,5 @@
 #
-# Shorewall-perl 4.0 -- /usr/share/shorewall-perl/Shorewall/Zones.pm
+# Shorewall-perl 4.2 -- /usr/share/shorewall-perl/Shorewall/Zones.pm
 #
 #     This program is under GPL [http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt]
 #
@@ -37,7 +37,6 @@ our @EXPORT = qw( NOTHING
 		  IPSECPROTO
 		  IPSECMODE
 
-		  numeric_value
 		  determine_zones
 		  zone_report
 		  dump_zone_contents
@@ -45,12 +44,15 @@ our @EXPORT = qw( NOTHING
 		  firewall_zone
 		  defined_zone
 		  zone_type
+		  zone_interfaces
 		  all_zones
 		  complex_zones
 		  non_firewall_zones
 		  single_interface
 		  validate_interfaces_file
 		  all_interfaces
+		  all_bridges
+		  interface_number
 		  find_interface
 		  known_interface
 		  have_bridges
@@ -59,12 +61,13 @@ our @EXPORT = qw( NOTHING
 		  interface_is_optional
 		  find_interfaces_by_option
 		  get_interface_option
+		  set_interface_option
 		  validate_hosts_file
 		  find_hosts_by_option
 		 );
 
 our @EXPORT_OK = qw( initialize );
-our $VERSION = 4.0.6;
+our $VERSION = 4.2.4;
 
 #
 # IPSEC Option types
@@ -81,21 +84,23 @@ use constant { NOTHING    => 'NOTHING',
 #
 #     @zones contains the ordered list of zones with sub-zones appearing before their parents.
 #
-#     %zones{<zone1> => {type = >      <zone type>       'firewall', 'ipv4', 'ipsec4', 'bport4';
+#     %zones{<zone1> => {type = >      <zone type>       'firewall', 'ip', 'ipsec', 'bport';
 #                        options =>    { complex => 0|1
+#                                        nested  => 0|1
 #                                        in_out  => < policy match string >
 #                                        in      => < policy match string >
 #                                        out     => < policy match string >
 #                                      }
-#                        parents =>    [ <parents> ]     Parents, Children and interfaces are listed by name
+#                        parents =>    [ <parents> ]      Parents, Children and interfaces are listed by name
 #                        children =>   [ <children> ]
-#                        interfaces => [ <interfaces> ]
+#                        interfaces => { <interfaces1> => 1, ... }
 #                        bridge =>     <bridge>
 #                        hosts { <type> } => [ { <interface1> => { ipsec   => 'ipsec'|'none'
 #                                                                  options => { <option1> => <value1>
 #                                                                               ...
 #                                                                             }
 #                                                                  hosts   => [ <net1> , <net2> , ... ]
+#                                                                  exclusions => [ <net1>, <net2>, ... ]
 #                                                                }
 #                                                <interface2> => ...
 #                                              }
@@ -120,19 +125,23 @@ our %reservedName = ( all => 1,
 #
 #     @interfaces lists the interface names in the order that they appear in the interfaces file.
 #
-#     %interfaces { <interface1> => { root        => <name without trailing '+'>
+#     %interfaces { <interface1> => { name        => <name of interface>
+#                                     root        => <name without trailing '+'>
 #                                     options     => { <option1> = <val1> ,
 #                                                      ...
 #                                                    }
 #                                     zone        => <zone name>
+#                                     nets        => <number of nets in interface/hosts records referring to this interface>
 #                                     bridge      => <bridge>
 #                                     broadcasts  => 'none', 'detect' or [ <addr1>, <addr2>, ... ]
+#                                     number      => <ordinal position in the interfaces file>
 #                                   }
 #                 }
 #
 our @interfaces;
 our %interfaces;
 our @bport_zones;
+our $family;
 
 #
 # Initialize globals -- we take this novel approach to globals initialization to allow
@@ -143,7 +152,8 @@ our @bport_zones;
 #                       the second and subsequent calls to that function.
 #
 
-sub initialize() {
+sub initialize( $ ) {
+    $family = shift;
     @zones = ();
     %zones = ();
     $firewall_zone = '';
@@ -154,16 +164,7 @@ sub initialize() {
 }
 
 INIT {
-    initialize;
-}
-
-#
-# Convert value to decimal number
-#
-sub numeric_value ( $ ) {
-    my $mark = lc $_[0];
-    fatal_error "Invalid Numeric Value ($mark)" unless $mark =~ /^(0x[a-f0-9]+|0[0-7]*|[1-9]\d*)$/;
-    $mark =~ /^0/ ? oct $mark : $mark;
+    initialize( F_IPV4 );
 }
 
 #
@@ -196,7 +197,7 @@ sub parse_zone_option_list($$)
     my $fmt;
 
     if ( $list ne '-' ) {
-	for my $e ( split ',' , $list ) {
+	for my $e ( split_list $list, 'option' ) {
 	    my $val    = undef;
 	    my $invert = '';
 
@@ -223,7 +224,7 @@ sub parse_zone_option_list($$)
 	    if ( $key{$e} ) {
 		$h{$e} = $val;
 	    } else {
-		fatal_error "The \"$e\" option may only be specified for ipsec zones" unless $zonetype eq 'ipsec4';
+		fatal_error "The \"$e\" option may only be specified for ipsec zones" unless $zonetype eq 'ipsec';
 		$options .= $invert;
 		$options .= "--$e ";
 		$options .= "$val "if defined $val;
@@ -243,7 +244,7 @@ sub determine_zones()
 {
     my @z;
 
-    my $ipv4 = 0;
+    my $ip = 0;
 
     my $fn = open_file 'zones';
 
@@ -257,7 +258,7 @@ sub determine_zones()
 
 	if ( $zone =~ /(\w+):([\w,]+)/ ) {
 	    $zone = $1;
-	    @parents = split ',', $2;
+	    @parents = split_list $2, 'zone';
 
 	    for my $p ( @parents ) {
 		fatal_error "Invalid Parent List ($2)" unless $p;
@@ -271,16 +272,23 @@ sub determine_zones()
 	fatal_error "Invalid zone name ($zone)"        if $reservedName{$zone} || $zone =~ /^all2|2all$/;
 	fatal_error( "Duplicate zone name ($zone)" ) if $zones{$zone};
 
-	$type = "ipv4" unless $type;
+	$type = "ip" unless $type;
 
 	if ( $type =~ /ipv4/i ) {
-	    $type = 'ipv4';
-	    $ipv4 = 1;
-	} elsif ( $type =~ /^ipsec4?$/i ) {
-	    $type = 'ipsec4';
-	} elsif ( $type =~ /^bport4?$/i ) {
+	    fatal_error "Invalid zone type ($type)" if $family == F_IPV6;
+	    $type = 'ip';
+	    $ip = 1;
+	} elsif ( $type =~ /ipv6/i ) {
+	    fatal_error "Invalid zone type ($type)" if $family == F_IPV4;
+	    $type = 'ip';
+	    $ip = 1;
+	} elsif ( $type =~ /^ipsec([46])?$/i ) {
+	    fatal_error "Invalid zone type ($type)" if $1 && (($1 == 4 && $family == F_IPV6 ) || ( $1 == 6 && $family == F_IPV4 ));
+	    $type = 'ipsec';
+	} elsif ( $type =~ /^bport([46])?$/i ) {
+	    fatal_error "Invalid zone type ($type)" if $1 && (( $1 == 4 && $family == F_IPV6 ) || ( $1 == 6 && $family == F_IPV4 ));
 	    warning_message "Bridge Port zones should have a parent zone" unless @parents;
-	    $type = 'bport4';
+	    $type = 'bport';
 	    push @bport_zones, $zone;
 	} elsif ( $type eq 'firewall' ) {
 	    fatal_error 'Firewall zone may not be nested' if @parents;
@@ -289,8 +297,8 @@ sub determine_zones()
 	    $ENV{FW} = $zone;
 	    $type = "firewall";
 	} elsif ( $type eq '-' ) {
-	    $type = 'ipv4';
-	    $ipv4 = 1;
+	    $type = 'ip';
+	    $ip = 1;
 	} else {
 	    fatal_error "Invalid zone type ($type)" ;
 	}
@@ -301,12 +309,12 @@ sub determine_zones()
 
 	$zones{$zone} = { type       => $type,
 			  parents    => \@parents,
-			  exclusions => [],
 			  bridge     => '',
 			  options    => { in_out  => parse_zone_option_list( $options || '', $type ) ,
 					  in      => parse_zone_option_list( $in_options || '', $type ) ,
 					  out     => parse_zone_option_list( $out_options || '', $type ) ,
-					  complex => ($type eq 'ipsec4' || $options || $in_options || $out_options ? 1 : 0) } ,
+					  complex => ($type eq 'ipsec' || $options || $in_options || $out_options ? 1 : 0) ,
+					  nested  => @parents > 0 } ,
 			  interfaces => {} ,
 			  children   => [] ,
 			  hosts      => {}
@@ -315,7 +323,7 @@ sub determine_zones()
     }
 
     fatal_error "No firewall zone defined" unless $firewall_zone;
-    fatal_error "No IPv4 zones defined" unless $ipv4;
+    fatal_error "No IP zones defined" unless $ip;
 
     my %ordered;
 
@@ -343,7 +351,7 @@ sub determine_zones()
 #
 sub haveipseczones() {
     for my $zoneref ( values %zones ) {
-	return 1 if $zoneref->{type} eq 'ipsec4';
+	return 1 if $zoneref->{type} eq 'ipsec';
     }
 
     0;
@@ -356,6 +364,8 @@ sub zone_report()
 {
     progress_message2 "Determining Hosts in Zones...";
 
+    my $ipzone = $family == F_IPV4 ? 'ipv4' : 'ipv6';
+
     for my $zone ( @zones )
     {
 	my $zoneref   = $zones{$zone};
@@ -363,7 +373,9 @@ sub zone_report()
 	my $type      = $zoneref->{type};
 	my $optionref = $zoneref->{options};
 
-	progress_message "   $zone ($type)";
+	$type = $ipzone if $type eq 'ip';
+
+	progress_message_nocompress "   $zone ($type)";
 
 	my $printed = 0;
 
@@ -374,10 +386,16 @@ sub zone_report()
 		for my $interface ( sort keys %$interfaceref ) {
 		    my $arrayref = $interfaceref->{$interface};
 		    for my $groupref ( @$arrayref ) {
-			my $hosts     = $groupref->{hosts};
+			my $hosts      = $groupref->{hosts};
+			my $exclusions = join ',', @{$groupref->{exclusions}};
 			if ( $hosts ) {
 			    my $grouplist = join ',', ( @$hosts );
-			    progress_message "      $interface:$grouplist";
+			    $grouplist = join '!', ( $grouplist, $exclusions) if $exclusions;
+			    if ( $family == F_IPV4 ) {
+				progress_message_nocompress "      $interface:$grouplist";
+			    } else {
+				progress_message_nocompress "      $interface:<$grouplist>";
+			    }
 			    $printed = 1;
 			}
 		    }
@@ -387,7 +405,7 @@ sub zone_report()
 	}
 
 	unless ( $printed ) {
-	    fatal_error "No bridge has been associated with zone $zone" if $type eq 'bport4' && ! $zoneref->{bridge};
+	    fatal_error "No bridge has been associated with zone $zone" if $type eq 'bport' && ! $zoneref->{bridge};
 	    warning_message "*** $zone is an EMPTY ZONE ***" unless $type eq 'firewall';
 	}
 
@@ -396,16 +414,30 @@ sub zone_report()
 
 sub dump_zone_contents()
 {
+    my %xlate;
+
+    if ( $family == F_IPV4 ) {
+	%xlate = ( ip =>    'ipv4' ,
+		   bport => 'bport4' ,
+		   ipsec => 'ipsec4' )
+    } else {
+	%xlate = ( ip =>    'ipv6' ,
+		   bport => 'bport6' ,
+		   ipsec => 'ipsec6' )
+    }
+
     for my $zone ( @zones )
     {
 	my $zoneref    = $zones{$zone};
 	my $hostref    = $zoneref->{hosts};
 	my $type       = $zoneref->{type};
 	my $optionref  = $zoneref->{options};
-	my $exclusions = $zoneref->{exclusions};
+
+	$type = $xlate{$type} if $xlate{$type};
+
 	my $entry      =  "$zone $type";
 
-	$entry .= ":$zoneref->{bridge}" if $type eq 'bport4';
+	$entry .= ":$zoneref->{bridge}" if $type =~ /^bport/;
 
 	if ( $hostref ) {
 	    for my $type ( sort keys %$hostref ) {
@@ -415,23 +447,24 @@ sub dump_zone_contents()
 		    my $arrayref = $interfaceref->{$interface};
 		    for my $groupref ( @$arrayref ) {
 			my $hosts     = $groupref->{hosts};
+			my $exclusions = join ',', @{$groupref->{exclusions}};
+
 			if ( $hosts ) {
 			    my $grouplist = join ',', ( @$hosts );
-			    $entry .= " $interface:$grouplist";
+
+			    $grouplist = join '!', ( $grouplist, $exclusions ) if $exclusions;
+
+			    if ( $family == F_IPV4 ) {
+				$entry .= " $interface:$grouplist";
+			    } else {
+				$entry .= " $interface:<$grouplist>";
+			    }
 			}
 		    }
 		}
 	    }
 	}
 
-	if ( @$exclusions ) {
-	    $entry .= ' exclude';
-
-	    for my $host ( @$exclusions ) {
-		$entry .= " $host";
-	    }
-	}
-
 	emit_unindented $entry;
     }
 }
@@ -442,7 +475,7 @@ sub dump_zone_contents()
 sub single_interface( $ ) {
     my $zone = $_[0];
     my $zoneref = $zones{$zone};
-  
+
     fatal_error "Internal Error in single_zone()" unless $zoneref;
 
     my @keys = keys( %{$zoneref->{interfaces}} );
@@ -453,9 +486,9 @@ sub single_interface( $ ) {
 sub add_group_to_zone($$$$$)
 {
     my ($zone, $type, $interface, $networks, $options) = @_;
+    my $hostsref;
     my $typeref;
     my $interfaceref;
-    my $arrayref;
     my $zoneref  = $zones{$zone};
     my $zonetype = $zoneref->{type};
     my $ifacezone = $interfaces{$interface}{zone};
@@ -463,13 +496,15 @@ sub add_group_to_zone($$$$$)
     $zoneref->{interfaces}{$interface} = 1;
 
     my @newnetworks;
-    my @exclusions;
+    my @exclusions = ();
     my $new = \@newnetworks;
     my $switched = 0;
 
     $ifacezone = '' unless defined $ifacezone;
 
     for my $host ( @$networks ) {
+	$interfaces{$interface}{nets}++;
+
 	fatal_error "Invalid Host List" unless defined $host and $host ne '';
 
 	if ( substr( $host, 0, 1 ) eq '!' ) {
@@ -482,32 +517,31 @@ sub add_group_to_zone($$$$$)
 	unless ( $switched ) {
 	    if ( $type eq $zonetype ) {
 		fatal_error "Duplicate Host Group ($interface:$host) in zone $zone" if $ifacezone eq $zone;
-		$ifacezone = $zone if $host eq ALLIPv4;
+		$ifacezone = $zone if $host eq ALLIP;
 	    }
 	}
 
 	if ( substr( $host, 0, 1 ) eq '+' ) {
 	    fatal_error "Invalid ipset name ($host)" unless $host =~ /^\+[a-zA-Z]\w*$/;
 	} else {
-	    validate_host $host, 1;
+	    validate_host $host, 0;
 	}
 
-	push @$new, $switched ? "$interface:$host" : $host;
+	push @$new, $host;
     }
 
     $zoneref->{options}{in_out}{routeback} = 1 if $options->{routeback};
 
-    $typeref      = ( $zoneref->{hosts}           || ( $zoneref->{hosts} = {} ) );
-    $interfaceref = ( $typeref->{$type}           || ( $interfaceref = $typeref->{$type} = {} ) );
-    $arrayref     = ( $interfaceref->{$interface} || ( $interfaceref->{$interface} = [] ) );
+    $hostsref     = ( $zoneref->{hosts}           || ( $zoneref->{hosts} = {} ) );
+    $typeref      = ( $hostsref->{$type}          || ( $hostsref->{$type} = {} ) );
+    $interfaceref = ( $typeref->{$interface}      || ( $typeref->{$interface} = [] ) );
 
-    $zoneref->{options}{complex} = 1 if @$arrayref || ( @newnetworks > 1 ) || ( @exclusions );
+    $zoneref->{options}{complex} = 1 if @$interfaceref || ( @newnetworks > 1 ) || ( @exclusions );
 
-    push @{$zoneref->{exclusions}}, @exclusions;
-
-    push @{$arrayref}, { options => $options,
-			 hosts   => \@newnetworks,
-			 ipsec   => $type eq 'ipsec4' ? 'ipsec' : 'none' };
+    push @{$interfaceref}, { options => $options,
+			     hosts   => \@newnetworks,
+			     ipsec   => $type eq 'ipsec' ? 'ipsec' : 'none' ,
+			     exclusions => \@exclusions };
 }
 
 #
@@ -528,6 +562,10 @@ sub zone_type( $ ) {
     find_zone( $_[0] )->{type};
 }
 
+sub zone_interfaces( $ ) {
+    find_zone( $_[0] )->{interfaces};
+}
+
 sub defined_zone( $ ) {
     $zones{$_[0]};
 }
@@ -555,6 +593,7 @@ sub firewall_zone() {
 sub validate_interfaces_file( $ )
 {
     my $export = shift;
+    my $num    = 0;
 
     use constant { SIMPLE_IF_OPTION   => 1,
 		   BINARY_IF_OPTION   => 2,
@@ -563,27 +602,47 @@ sub validate_interfaces_file( $ )
 		   OBSOLETE_IF_OPTION => 5,
 	           MASK_IF_OPTION     => 7,
 
-	           IF_OPTION_ZONEONLY => 8 };
+	           IF_OPTION_ZONEONLY => 8,
+	           IF_OPTION_HOST     => 16,
+	       };
 
-    my %validoptions = (arp_filter  => BINARY_IF_OPTION,
-			arp_ignore  => ENUM_IF_OPTION,
-			blacklist   => SIMPLE_IF_OPTION,
-			bridge      => SIMPLE_IF_OPTION,
-			detectnets  => OBSOLETE_IF_OPTION,
-			dhcp        => SIMPLE_IF_OPTION,
-			maclist     => SIMPLE_IF_OPTION,
-			logmartians => BINARY_IF_OPTION,
-			norfc1918   => SIMPLE_IF_OPTION,
-			nosmurfs    => SIMPLE_IF_OPTION,
-			optional    => SIMPLE_IF_OPTION,
-			proxyarp    => BINARY_IF_OPTION,
-			routeback   => SIMPLE_IF_OPTION + IF_OPTION_ZONEONLY,
-			routefilter => BINARY_IF_OPTION,
-			sourceroute => BINARY_IF_OPTION,
-			tcpflags    => SIMPLE_IF_OPTION,
-			upnp        => SIMPLE_IF_OPTION,
-			mss         => NUMERIC_IF_OPTION,
+    my %validoptions;
+
+    if ( $family == F_IPV4 ) {
+	%validoptions = (arp_filter  => BINARY_IF_OPTION,
+			 arp_ignore  => ENUM_IF_OPTION,
+			 blacklist   => SIMPLE_IF_OPTION + IF_OPTION_HOST,
+			 bridge      => SIMPLE_IF_OPTION,
+			 detectnets  => OBSOLETE_IF_OPTION,
+			 dhcp        => SIMPLE_IF_OPTION,
+			 maclist     => SIMPLE_IF_OPTION + IF_OPTION_HOST,
+			 logmartians => BINARY_IF_OPTION,
+			 norfc1918   => SIMPLE_IF_OPTION + IF_OPTION_HOST,
+			 nosmurfs    => SIMPLE_IF_OPTION + IF_OPTION_HOST,
+			 optional    => SIMPLE_IF_OPTION,
+			 proxyarp    => BINARY_IF_OPTION,
+			 routeback   => SIMPLE_IF_OPTION + IF_OPTION_ZONEONLY + IF_OPTION_HOST,
+			 routefilter => BINARY_IF_OPTION,
+			 sourceroute => BINARY_IF_OPTION,
+			 tcpflags    => SIMPLE_IF_OPTION + IF_OPTION_HOST,
+			 upnp        => SIMPLE_IF_OPTION,
+			 mss         => NUMERIC_IF_OPTION,
 			);
+    } else {
+	%validoptions = (  blacklist   => SIMPLE_IF_OPTION + IF_OPTION_HOST,
+			   bridge      => SIMPLE_IF_OPTION,
+			   dhcp        => SIMPLE_IF_OPTION,
+			   maclist     => SIMPLE_IF_OPTION + IF_OPTION_HOST,
+			   nosmurfs    => SIMPLE_IF_OPTION,
+			   optional    => SIMPLE_IF_OPTION,
+			   proxyndp    => BINARY_IF_OPTION,
+ 			   routeback   => SIMPLE_IF_OPTION + IF_OPTION_ZONEONLY + IF_OPTION_HOST,
+			   sourceroute => BINARY_IF_OPTION,
+			   tcpflags    => SIMPLE_IF_OPTION + IF_OPTION_HOST,
+			   mss         => NUMERIC_IF_OPTION,
+			   forward     => NUMERIC_IF_OPTION,
+			  );
+    }
 
     my $fn = open_file 'interfaces';
 
@@ -616,9 +675,7 @@ sub validate_interfaces_file( $ )
 
 	my ($interface, $port, $extra) = split /:/ , $originalinterface, 3;
 
-	fatal_error "Invalid INTERFACE ($originalinterface)" if defined $extra || ! $interface;
-
-	fatal_error "Invalid Interface Name ($interface)" if $interface eq '+';
+	fatal_error "Invalid INTERFACE ($originalinterface)" if ! $interface || defined $extra;
 
 	if ( defined $port ) {
 	    fatal_error qq("Virtual" interfaces are not supported -- see http://www.shorewall.net/Shorewall_and_Aliased_Interfaces.html) if $port =~ /^\d+$/;
@@ -626,7 +683,7 @@ sub validate_interfaces_file( $ )
 	    fatal_error "Your iptables is not recent enough to support bridge ports" unless $capabilities{KLUDGEFREE};
 	    fatal_error "Duplicate Interface ($port)" if $interfaces{$port};
 	    fatal_error "$interface is not a defined bridge" unless $interfaces{$interface} && $interfaces{$interface}{options}{bridge};
-	    fatal_error "Bridge Ports may only be associated with 'bport' zones" if $zone && $zoneref->{type} ne 'bport4';
+	    fatal_error "Bridge Ports may only be associated with 'bport' zones" if $zone && $zoneref->{type} ne 'bport';
 
 	    if ( $zone ) {
 		if ( $zoneref->{bridge} ) {
@@ -636,31 +693,32 @@ sub validate_interfaces_file( $ )
 		}
 	    }
 
-	    fatal_error "Bridge Ports may not have options" if $options && $options ne '-';
-
 	    next if $port eq '';
 
 	    fatal_error "Invalid Interface Name ($interface:$port)" unless $port =~ /^[\w.@%-]+\+?$/;
 
-	    $interfaces{$port}{bridge} = $bridge = $interface;
+	    $bridge = $interface;
 	    $interface = $port;
 	} else {
 	    fatal_error "Duplicate Interface ($interface)" if $interfaces{$interface};
-	    fatal_error "Zones of type 'bport' may only be associated with bridge ports" if $zone && $zoneref->{type} eq 'bport4';
-	    $interfaces{$interface}{bridge} = $interface;
+	    fatal_error "Zones of type 'bport' may only be associated with bridge ports" if $zone && $zoneref->{type} eq 'bport';
+	    $bridge = $interface;
 	}
 
 	my $wildcard = 0;
+	my $root;
 
 	if ( $interface =~ /\+$/ ) {
 	    $wildcard = 1;
-	    $interfaces{$interface}{root} = substr( $interface, 0, -1 );
+	    $root = substr( $interface, 0, -1 );
 	} else {
-	    $interfaces{$interface}{root} = $interface;
+	    $root = $interface;
 	}
 
+	my $broadcasts;
+
 	unless ( $networks eq '' || $networks eq 'detect' ) {
-	    my @broadcasts = split /,/, $networks;
+	    my @broadcasts = split_list $networks, 'address';
 
 	    for my $address ( @broadcasts ) {
 		fatal_error 'Invalid BROADCAST address' unless $address =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/;
@@ -669,7 +727,7 @@ sub validate_interfaces_file( $ )
 	    if ( $capabilities{ADDRTYPE} ) {
 		warning_message 'Shorewall no longer uses broadcast addresses in rule generation when Address Type Match is available';
 	    } else {
-		$interfaces{$interface}{broadcasts} = \@broadcasts;
+		$broadcasts = \@broadcasts;
 	    }
 	}
 
@@ -677,9 +735,11 @@ sub validate_interfaces_file( $ )
 
 	my %options;
 
+	$options{port} = 1 if $port;
+
 	if ( $options ) {
 
-	    for my $option (split ',', $options ) {
+	    for my $option (split_list $options, 'option' ) {
 		next if $option eq '-';
 
 		( $option, my $value ) = split /=/, $option;
@@ -688,6 +748,10 @@ sub validate_interfaces_file( $ )
 
 		fatal_error "The \"$option\" option may not be specified on a multi-zone interface" if $type & IF_OPTION_ZONEONLY && ! $zone;
 
+		my $hostopt = $type & IF_OPTION_HOST;
+
+		fatal_error "The \"$option\" option is not allowed on a bridge port" if $port && ! $hostopt;
+
 		$type &= MASK_IF_OPTION;
 
 		if ( $type == SIMPLE_IF_OPTION ) {
@@ -715,7 +779,9 @@ sub validate_interfaces_file( $ )
 		    }
 		} elsif ( $type == NUMERIC_IF_OPTION ) {
 		    fatal_error "The $option option requires a value" unless defined $value;
-		    $options{$option} = numeric_value $value;
+		    my $numval = numeric_value $value;
+		    fatal_error "Invalid value ($value) for option $option" unless defined $numval;
+		    $options{$option} = $numval;
 		} else {
 		    warning_message "Support for the $option interface option has been removed from Shorewall-perl";
 		}
@@ -727,21 +793,27 @@ sub validate_interfaces_file( $ )
 		require_capability( 'PHYSDEV_MATCH', 'The "bridge" option', 's');
 		fatal_error "Bridges may not have wildcard names" if $wildcard;
 	    }
-	} elsif ( $port ) {
-	    $options{port} = 1;
 	}
 
-	$interfaces{$interface}{options} = $optionsref = \%options;
+	$optionsref = \%options;
+
+	$interfaces{$interface} = { name       => $interface ,
+				    bridge     => $bridge ,
+				    nets       => 0 ,
+				    number     => ++$num ,
+				    root       => $root ,
+				    broadcasts => $broadcasts ,
+				    options    => $optionsref };
 
 	push @ifaces, $interface;
 
-	my @networks = allipv4;
+	my @networks = allip;
 
-	add_group_to_zone( $zone, $zoneref->{type}, $interface, \@networks, $optionsref ) if $zone && @networks;
+	add_group_to_zone( $zone, $zoneref->{type}, $interface, \@networks, $optionsref ) if $zone;
 
     	$interfaces{$interface}{zone} = $zone; #Must follow the call to add_group_to_zone()
 
-	progress_message "   Interface \"$currentline\" Validated";
+	progress_message "  Interface \"$currentline\" Validated";
 
     }
 
@@ -777,25 +849,32 @@ sub validate_interfaces_file( $ )
 sub known_interface($)
 {
     my $interface = $_[0];
+    my $interfaceref = $interfaces{$interface};
 
-    return 1 if $interfaces{$interface};
+    return $interfaceref if $interfaceref;
 
     for my $i ( @interfaces ) {
-	my $interfaceref = $interfaces{$i};
+	$interfaceref = $interfaces{$i};
 	my $val = $interfaceref->{root};
 	next if $val eq $i;
 	if ( substr( $interface, 0, length $val ) eq $val ) {
 	    #
-	    # Cache this result for future reference
+	    # Cache this result for future reference. We set the 'name' to the name of the entry that appears in /etc/shorewall/interfaces.
 	    #
-	    $interfaces{$interface} = { options => $interfaceref->{options}, bridge => $interfaceref->{bridge} };
-	    return 1;
+	    return $interfaces{$interface} = { options => $interfaceref->{options}, bridge => $interfaceref->{bridge} , name => $i , number => $interfaceref->{number} };
 	}
     }
 
     0;
 }
 
+#
+# Return interface number
+#
+sub interface_number( $ ) {
+    $interfaces{$_[0]}{number} || 256;
+}
+
 #
 # Return the interfaces list
 #
@@ -803,13 +882,20 @@ sub all_interfaces() {
     @interfaces;
 }
 
+#
+# Return a list of bridges
+#
+sub all_bridges() {
+    grep ( $interfaces{$_}{options}{bridge} , @interfaces );
+}
+
 #
 # Return a reference to the interfaces table entry for an interface
 #
 sub find_interface( $ ) {
     my $interface    = $_[0];
     my $interfaceref = $interfaces{ $interface };
-    
+
     fatal_error "Unknown Interface ($interface)" unless $interfaceref;
 
     $interfaceref;
@@ -873,23 +959,42 @@ sub get_interface_option( $$ ) {
     $interfaces{$interface}{options}{$option};
 }
 
+#
+# Set an option for an interface
+#
+sub set_interface_option( $$$ ) {
+    my ( $interface, $option, $value ) = @_;
+
+    $interfaces{$interface}{options}{$option} = $value;
+}
+
 #
 # Validates the hosts file. Generates entries in %zone{..}{hosts}
 #
 sub validate_hosts_file()
 {
-    my %validoptions = (
-			blacklist => 1,
-			maclist => 1,
-			norfc1918 => 1,
-			nosmurfs => 1,
-			routeback => 1,
-			routefilter => 1,
-			tcpflags => 1,
-			broadcast => 1,
-			destonly => 1,
-			sourceonly => 1,
+    my %validoptions;
+
+    if ( $family == F_IPV4 ) {
+	%validoptions = (
+			 blacklist => 1,
+			 maclist => 1,
+			 norfc1918 => 1,
+			 nosmurfs => 1,
+			 routeback => 1,
+			 tcpflags => 1,
+			 broadcast => 1,
+			 destonly => 1,
+			 sourceonly => 1,
 			);
+    } else {
+	%validoptions = (
+			 blacklist => 1,
+			 maclist => 1,
+			 routeback => 1,
+			 tcpflags => 1,
+			);
+    }
 
     my $ipsec = 0;
     my $first_entry = 1;
@@ -913,16 +1018,27 @@ sub validate_hosts_file()
 
 	my $interface;
 
-	if ( $hosts =~ /^([\w.@%-]+\+?):(.*)$/ ) {
-	    $interface = $1;
-	    $hosts = $2;
-	    $zoneref->{options}{complex} = 1 if $hosts =~ /^\+/;
-	    fatal_error "Unknown interface ($interface)" unless $interfaces{$interface}{root};
+	if ( $family == F_IPV4 ) {
+	    if ( $hosts =~ /^([\w.@%-]+\+?):(.*)$/ ) {
+		$interface = $1;
+		$hosts = $2;
+		$zoneref->{options}{complex} = 1 if $hosts =~ /^\+/;
+		fatal_error "Unknown interface ($interface)" unless $interfaces{$interface}{root};
+	    } else {
+		fatal_error "Invalid HOST(S) column contents: $hosts";
+	    }
 	} else {
-	    fatal_error "Invalid HOST(S) column contents: $hosts";
+	    if ( $hosts =~ /^([\w.@%-]+\+?):<(.*)>\s*$/ ) {
+		$interface = $1;
+		$hosts = $2;
+		$zoneref->{options}{complex} = 1 if $hosts =~ /^\+/;
+		fatal_error "Unknown interface ($interface)" unless $interfaces{$interface}{root};
+	    } else {
+		fatal_error "Invalid HOST(S) column contents: $hosts";
+	    }
 	}
 
-	if ( $type eq 'bport4' ) {
+	if ( $type eq 'bport' ) {
 	    if ( $zoneref->{bridge} eq '' ) {
 		fatal_error 'Bridge Port Zones may only be associated with bridge ports' unless $interfaces{$interface}{options}{port};
 		$zoneref->{bridge} = $interfaces{$interface}{bridge};
@@ -934,13 +1050,13 @@ sub validate_hosts_file()
 	my $optionsref = {};
 
 	if ( $options ne '-' ) {
-	    my @options = split ',', $options;
+	    my @options = split_list $options, 'option';
 	    my %options;
 
 	    for my $option ( @options )
 	    {
 		if ( $option eq 'ipsec' ) {
-		    $type = 'ipsec4';
+		    $type = 'ipsec';
 		    $zoneref->{options}{complex} = 1;
 		    $ipsec = 1;
 		} elsif ( $validoptions{$option}) {
@@ -966,9 +1082,9 @@ sub validate_hosts_file()
 	#
 	# Take care of case where the hosts list begins with '!'
 	#
-	$hosts = join( '', ALLIPv4 , $hosts ) if substr($hosts, 0, 2 ) eq ',!';
+	$hosts = join( '', ALLIP , $hosts ) if substr($hosts, 0, 2 ) eq ',!';
 
-	add_group_to_zone( $zone, $type , $interface, [ split( ',', $hosts ) ] , $optionsref);
+	add_group_to_zone( $zone, $type , $interface, [ split_list( $hosts, 'host' ) ] , $optionsref);
 
 	progress_message "   Host \"$currentline\" validated";
     }
@@ -978,7 +1094,7 @@ sub validate_hosts_file()
 
 #
 # Returns a reference to a array of host entries. Each entry is a
-# reference to an array containing ( interface , polciy match type {ipsec|none} , network );
+# reference to an array containing ( interface , polciy match type {ipsec|none} , network , exclusions );
 #
 sub find_hosts_by_option( $ ) {
     my $option = $_[0];
@@ -990,7 +1106,7 @@ sub find_hosts_by_option( $ ) {
 		for my $host ( @{$arrayref} ) {
 		    if ( $host->{options}{$option} ) {
 			for my $net ( @{$host->{hosts}} ) {
-			    push @hosts, [ $interface, $host->{ipsec} , $net ];
+			    push @hosts, [ $interface, $host->{ipsec} , $net , $host->{exclusions}];
 			}
 		    }
 		}
@@ -1000,7 +1116,7 @@ sub find_hosts_by_option( $ ) {
 
     for my $interface ( @interfaces ) {
 	if ( ! $interfaces{$interface}{zone} && $interfaces{$interface}{options}{$option} ) {
-	    push @hosts, [ $interface, 'none', ALLIPv4 ];
+	    push @hosts, [ $interface, 'none', ALLIP , [] ];
 	}
     }
 
